from __future__ import annotations

from enum import Enum
from typing import Any, List, Optional, Union

from pydantic import Field

from iambic.core.models import BaseModel, ExpiryModel

# Reference: https://www.guidodiepen.nl/2019/02/implementing-a-simple-plugin-framework-in-python/


class IdentityProvider(BaseModel):
    name: str


class UserStatus(Enum):
    active = "active"
    provisioned = "provisioned"
    deprovisioned = "deprovisioned"
    recovery = "recovery"
    suspended = "suspended"
    staged = "staged"
    locked_out = "locked_out"
    password_expired = "password_expired"


class User(BaseModel, ExpiryModel):
    idp_name: str
    username: str
    user_id: Optional[str]
    domain: Optional[str]
    fullname: Optional[str]
    status: Optional[UserStatus]
    created: Optional[str]
    updated: Optional[str]
    groups: Optional[List[str]]
    background_check_status: Optional[bool]
    extra: Any = Field(None, description=("Extra attributes to store"))
    profile: dict[str, Any]

    @property
    def resource_type(self) -> str:
        return "okta:user"

    @property
    def resource_id(self) -> str:
        return self.username


class GroupAttributes(BaseModel):
    requestable: bool = Field(
        False, description="Whether end-users can request access to group"
    )
    manager_approval_required: bool = Field(
        False, description="Whether a manager needs to approve access to the group"
    )
    approval_chain: List[Union[User, str]] = Field(
        [],
        description="A list of users or groups that need to approve access to the group",
    )
    self_approval_groups: List[str] = Field(
        [],
        description=(
            "If the user is a member of a self-approval group, their request to the group "
            "will be automatically approved"
        ),
    )
    allow_bulk_add_and_remove: bool = Field(
        True,
        description=(
            "Controls whether administrators can automatically approve access to the group"
        ),
    )
    background_check_required: bool = Field(
        False,
        description=("Whether a background check is required to be added to the group"),
    )
    allow_contractors: bool = Field(
        False,
        description=("Whether contractors are allowed to be members of the group"),
    )
    allow_third_party: bool = Field(
        False,
        description=(
            "Whether third-party users are allowed to be a member of the group"
        ),
    )
    emails_to_notify_on_new_members: List[str] = Field(
        [],
        description=(
            "A list of e-mail addresses to notify when new users are added to the group."
        ),
    )


class Group(BaseModel):
    name: str = Field(..., description="Name of the group")
    owner: Optional[str] = Field(None, description="Owner of the group")
    idp_name: str = Field(
        ...,
        description="Name of the tenant's identity provider that's associated with the group",
    )
    group_id: Optional[str] = Field(
        ..., description="Unique Group ID for the group. Usually it's {idp-name}-{name}"
    )
    description: Optional[str] = Field(None, description="Description of the group")
    attributes: GroupAttributes = Field(
        ...,
        description=(
            "Protected attributes that tell us whether the group is requester, where approvals should be routed, etc."
        ),
    )
    extra: Any = Field(None, description=("Extra attributes to store"))
    members: List[User] = Field([], description="Users in the group")

    @property
    def resource_type(self) -> str:
        return "okta:group"


class Assignment(BaseModel, ExpiryModel):
    user: Optional[str] = Field(None, description="User assigned to the app")
    group: Optional[str] = Field(None, description="Group assigned to the app")

    @property
    def resource_type(self) -> str:
        return "okta:group:assignment"

    @property
    def resource_id(self) -> str:
        return f"{self.user or self.group}"


class Status(Enum):
    active = "ACTIVE"
    inactive = "INACTIVE"


class AppProfileMapping(BaseModel):
    id: str
    name: str
    source: str
    target: str


class App(BaseModel, ExpiryModel):
    id: str = Field(..., description="ID of the app")
    idp_name: str = Field(..., description="Name of the identity provider")
    name: str = Field(..., description="Name of the app")
    status: Optional[Status] = Field(None, description="Status of the app")
    created: Optional[str] = Field(None, description="Date the app was created")
    last_updated: Optional[str] = Field(
        None, description="Date the app was last updated"
    )
    accessibility: Optional[dict] = Field(None, description="Accessibility settings")
    visibility: Optional[dict] = Field(None, description="Visibility settings")
    features: Optional[list] = Field(None, description="Features settings")
    sign_on_mode: Optional[str] = Field(None, description="Sign-on mode")
    credentials: Optional[dict] = Field(None, description="Credentials settings")
    settings: Optional[dict] = Field(None, description="Settings")
    assignments: list[Assignment] = Field([], description="Assignments")
    profile_mappings: list[AppProfileMapping] = Field(
        [], description="Profile mappings"
    )
    attributes: Optional[dict] = Field(None, description="Attributes", exclude=True)
    extra: Any = Field(None, description=("Extra attributes to store"), exclude=True)

    @property
    def resource_type(self) -> str:
        return "okta:app"

    @property
    def resource_id(self) -> str:
        return f"{self.id}"


class ActionStatus(Enum):
    success = "success"
    error = "error"


class ActionResponse(BaseModel):
    status: Optional[ActionStatus] = None
    errors: Optional[List[str]] = None
    data: Any = None


class GroupRequestStatus(Enum):
    pending = "pending"
    approved = "approved"
    cancelled = "cancelled"
    rejected = "rejected"


class LastUpdated(BaseModel):
    user: User
    time: int
    comment: str


class GroupRequest(BaseModel):
    request_id: str
    request_url: str
    tenant: str
    users: List[User]
    groups: List[Group]
    requester: User
    justification: str
    expires: Optional[int] = None
    status: GroupRequestStatus
    created_time: int
    last_updated: List[LastUpdated]
    last_updated_time: int
    last_updated_by: User
    reviewer_comments: Optional[str]


class GroupRequests(BaseModel):
    requests: List[GroupRequest]


class GroupRequestsTable(BaseModel):
    User: str
    Group: str
    Requester: str
    Justification: str
    Expires: Optional[str]
    Status: str
    Last_Updated: str



from __future__ import annotations

import asyncio


async def handle_azure_ad_fn(fn, *args, **kwargs):
    try:
        res = await fn(*args, **kwargs)
    except asyncio.exceptions.TimeoutError:
        raise asyncio.exceptions.TimeoutError
    return res



from __future__ import annotations

import functools
import secrets
from typing import TYPE_CHECKING, List, Optional

from aiohttp import ClientResponseError

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import GlobalRetryController, snake_to_camelback
from iambic.plugins.v0_1_0.azure_ad.user.models import UserTemplateProperties

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.azure_ad.models import AzureADOrganization


async def list_users(
    azure_ad_organization: AzureADOrganization, **kwargs
) -> List[UserTemplateProperties]:
    """
    List all users in Azure AD.

    Args:
    - azure_ad_organization: An instance of the AzureADOrganization class, which provides access to the Azure AD API.

    Returns:
    - A list of `User` instances, representing the users in Azure AD.
    """

    async with GlobalRetryController(
        fn_identifier="azure_ad.list_users"
    ) as retry_controller:
        fn = functools.partial(azure_ad_organization.list, "users", **kwargs)
        users = await retry_controller(fn)

    return [UserTemplateProperties.from_azure_response(user) for user in users]


async def get_user(
    azure_ad_organization: AzureADOrganization,
    user_id: Optional[str] = None,
    username: Optional[str] = None,
    allow_template_ref: bool = False,
) -> UserTemplateProperties:
    """
    Get Azure AD user.

    Args:
    - azure_ad_organization: An instance of the AzureADOrganization class, which provides access to the Azure AD API.
    - user_id: The user ID to get.
    - allow_template_ref: If True, attempt to resolve the user by loading the IAMbic template

    Returns:
    - The `User` instance, representing the user in Azure AD.
    """
    assert user_id or username

    if user_id:
        async with GlobalRetryController(
            fn_identifier="azure_ad.get_user"
        ) as retry_controller:
            fn = functools.partial(azure_ad_organization.get, f"users/{user_id}")
            user = await retry_controller(fn)
            return UserTemplateProperties.from_azure_response(user)

    users = await list_users(
        azure_ad_organization,
        params={"$filter": f"userPrincipalName eq '{username}'"},
    )
    if users:
        return users[0]

    raise Exception(f"User not found with username {username}")


async def create_user(
    azure_ad_organization: AzureADOrganization,
    username: str,
    mail_nickname: Optional[str],
    display_name: str,
) -> Optional[UserTemplateProperties]:
    """
    Create a new user in Azure AD.
    Args:
    - azure_ad_organization (AzureADOrganization): The Azure AD organization to update the user in.
    - username: The name of the user to create.
    - mail_nickname: The mail alias for the user.
    - display_name: The name to display in the address book for the user.

    Returns:
    - An instance of the `UserTemplateProperties` class, representing the created user.
    """

    if ctx.execute:
        log.warning(
            "request data",
            request_data={
                "accountEnabled": True,
                "displayName": display_name,
                "mailNickname": mail_nickname,
                "userPrincipalName": username,
            },
        )

        user = await azure_ad_organization.post(
            "users",
            json={
                "accountEnabled": True,
                "displayName": display_name,
                "mailNickname": mail_nickname,
                "userPrincipalName": username,
                "passwordProfile": {
                    "forceChangePasswordNextSignIn": True,
                    "forceChangePasswordNextSignInWithMfa": azure_ad_organization.require_user_mfa_on_create,
                    "password": secrets.token_urlsafe(15),
                },
            },
        )
        return UserTemplateProperties.from_azure_response(user)


async def update_user_attributes(
    azure_ad_organization: AzureADOrganization,
    template_user: UserTemplateProperties,
    cloud_user: UserTemplateProperties,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the name of a user in Azure AD.
        Args:
        azure_ad_organization (AzureADOrganization): The Azure AD organization to update the user in.
        template_user (UserTemplateProperties): The template representation of the user.
        cloud_user (UserTemplateProperties): The current representation of the user in the cloud.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = []
    patch_request = {}

    for attr, value in cloud_user.dict(
        exclude_none=False, exclude={"user_id", "fullname"}
    ).items():
        if (template_value := getattr(template_user, attr)) != value:
            response.append(
                ProposedChange(
                    change_type=ProposedChangeType.UPDATE,
                    resource_id=template_user.user_id,
                    resource_type=template_user.resource_type,
                    attribute=attr,
                    current_value=value,
                    new_value=template_value,
                )
            )

            attr = (
                "username" if attr == "userPrincipalName" else snake_to_camelback(attr)
            )
            patch_request[attr] = template_value

    if ctx.execute and patch_request:
        try:
            await azure_ad_organization.patch(
                f"users/{cloud_user.user_id}",
                json=patch_request,
            )
        except ClientResponseError as err:
            log.exception(
                "Failed to update user in Azure AD",
                **log_params,
            )
            response[0].exceptions_seen = [str(err)]

    return response


async def delete_user(
    azure_ad_organization: AzureADOrganization,
    user: UserTemplateProperties,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Delete a user in Azure AD.
    Args:
        azure_ad_organization (AzureADOrganization): The Azure AD organization to delete the user from.
        user (UserTemplateProperties): The user to delete.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = [
        ProposedChange(
            change_type=ProposedChangeType.DELETE,
            resource_id=user.user_id,
            resource_type=user.resource_type,
            attribute="user",
            change_summary={"user": user.username},
            current_value=user.username,
            new_value=None,
        )
    ]

    if ctx.execute:
        try:
            await azure_ad_organization.delete(f"users/{user.user_id}")
        except ClientResponseError as err:
            log.exception(
                "Failed to delete user in Azure AD",
                **log_params,
            )
            response[0].exceptions_seen = [str(err)]

    return response



from __future__ import annotations

from io import StringIO
from typing import List

from rich.console import Console
from rich.style import Style
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from iambic.output.models import ActionSummary, ExceptionSummary

console = Console()


def rich_format(value: str, style_str: str) -> str:
    style = Style.parse(style_str)
    return console.render_str(str(value), style=style)


def rich_text(text: str, style_str: str) -> str:
    style = Style.parse(style_str)
    return Text(str(text), style=style)


def rich_text_table(table_headers: List[str], table_data: List[List[str]]) -> str:
    table = Table(show_header=True, header_style="bold magenta")
    for header in table_headers:
        table.add_column(header, justify="left", style="dim", no_wrap=True)
    for row in table_data:
        table.add_row(*row)
    console.render_lines(table)


def rich_tree_summary(action_summary: ActionSummary):
    action_summary_tree = Tree(action_summary.action, expanded=True)
    for template in action_summary.templates:
        template_tree = action_summary_tree.add(template.template_path, expanded=True)
        for account in template.accounts:
            account_tree = template_tree.add(account.account, expanded=True)
            for change in account.changes:
                change_tree = account_tree.add(
                    f"{change.change.resource_type} // {change.change.resource_id}",
                    expanded=True,
                )
                if change.change.diff:
                    change_tree.add(change.change.diff_plus_minus)
    console = Console(file=StringIO(), force_terminal=True)
    console.print(action_summary_tree)
    output = console.file.getvalue()
    return output


def rich_tree_exception(exceptions: ExceptionSummary):
    exception_tree = Tree(exceptions.action, expanded=True)
    for template in exceptions.templates:
        template_tree = exception_tree.add(template.template_path, expanded=True)
        for account in template.accounts:
            account_tree = template_tree.add(account.account, expanded=True)
            for change in account.changes:
                change_tree = account_tree.add(
                    str(change.change.change_type), expanded=True
                )
                change_tree.add(change.change.resource_id)
                change_tree.add(change.change.resource_type)
                change_tree.add(str(change.change.change_type.value))
                if change.change.diff:
                    change_tree.add("* " + "\n* ".join(change.change.diff_plus_minus))
    console = Console(file=StringIO(), force_terminal=True)
    console.print(exception_tree)
    output = console.file.getvalue()
    return output



from __future__ import annotations

import os
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, MagicMock, patch

from pydantic import SecretStr

from iambic.core.iambic_enum import IambicManaged
from iambic.core.models import ExecutionMessage
from iambic.plugins.v0_1_0.google_workspace.group.models import (
    GoogleWorkspaceGroupTemplate,
)
from iambic.plugins.v0_1_0.google_workspace.group.template_generation import (
    collect_project_groups,
    generate_domain_group_resource_files,
    generate_group_templates,
    get_group_dir,
    get_resource_dir_args,
    get_response_dir,
    get_templated_resource_file_path,
    update_or_create_group_template,
)
from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import (
    GoogleProject,
    GoogleSubject,
    GoogleWorkspaceConfig,
)


class TestGenerateGroupTemplates(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.exe_message = MagicMock(spec=ExecutionMessage)
        self.exe_message.get_sub_exe_files = AsyncMock(return_value=[])
        self.config = GoogleWorkspaceConfig(
            workspaces=[
                GoogleProject(
                    project_id="example-ga",
                    project_name=None,
                    subjects=[
                        GoogleSubject(
                            domain="example.com",
                            service_account="google-groups-test@example-ga.iam.gserviceaccount.com",
                        )
                    ],
                    type="service_account",
                    private_key_id="12345",
                    private_key=SecretStr("**********"),
                    client_email="google-groups-test@example-ga.iam.gserviceaccount.com",
                    client_id="12345",
                    auth_uri="https://accounts.google.com/o/oauth2/auth",
                    token_uri="https://oauth2.googleapis.com/token",
                    auth_provider_x509_cert_url="https://www.googleapis.com/oauth2/v1/certs",
                    client_x509_cert_url="https://www.googleapis.com/robot/v1/metadata/x509/google-groups-test%40example-ga.iam.gserviceaccount.com",
                    variables=[],
                    iambic_managed=IambicManaged.UNDEFINED,
                )
            ]
        )
        self.output_dir = "test_output"

    @patch(
        "iambic.plugins.v0_1_0.google_workspace.group.template_generation.get_existing_template_map",
        return_value={},
    )
    @patch(
        "iambic.plugins.v0_1_0.google_workspace.group.template_generation.update_or_create_group_template"
    )
    @patch(
        "iambic.plugins.v0_1_0.google_workspace.group.template_generation.delete_orphaned_templates"
    )
    async def test_generate_group_templates_no_groups(
        self,
        mock_delete_orphaned_templates,
        mock_update_or_create_group_template,
        mock_get_existing_template_map,
    ):
        await generate_group_templates(
            self.exe_message,
            self.config,
            self.output_dir,
        )

        self.exe_message.get_sub_exe_files.assert_called_once()
        mock_get_existing_template_map.assert_called_once()
        mock_update_or_create_group_template.assert_not_called()
        mock_delete_orphaned_templates.assert_called_once_with([], set())

    async def test_generate_group_templates_with_groups(self):
        groups = [
            {
                "properties": {
                    "name": "Group 1",
                    "email": "group1@example.com",
                    "description": "Group 1 description",
                    "domain": "example.com",
                    "members": [],
                },
            },
        ]

        group_template = GoogleWorkspaceGroupTemplate(
            **groups[0], file_path="test_path.yaml"
        )

        self.exe_message.get_sub_exe_files = AsyncMock(return_value=groups)

        with patch(
            "iambic.plugins.v0_1_0.google_workspace.group.template_generation.get_existing_template_map",
            return_value={},
        ), patch(
            "iambic.plugins.v0_1_0.google_workspace.group.template_generation.update_or_create_group_template"
        ) as mock_update_or_create_group_template, patch(
            "iambic.plugins.v0_1_0.google_workspace.group.template_generation.delete_orphaned_templates"
        ) as mock_delete_orphaned_templates:
            mock_update_or_create_group_template.return_value = group_template
            await generate_group_templates(
                self.exe_message,
                self.config,
                self.output_dir,
            )

            self.exe_message.get_sub_exe_files.assert_called_once()
            mock_update_or_create_group_template.assert_called_once()
            mock_delete_orphaned_templates.assert_called_once_with(
                [], {"group1@example.com"}
            )


class TestTemplateGenerationFunctions(IsolatedAsyncioTestCase):
    def test_get_resource_dir_args(self):
        domain = "example.com"
        expected_args = ["group", "example.com"]
        self.assertEqual(get_resource_dir_args(domain), expected_args)

    def test_get_response_dir(self):
        exe_message = ExecutionMessage(
            provider_id="test_provider_id",
            execution_id="test_execution_id",
            command="import",
        )
        # exe_message.get_directory.return_value = "test_directory"
        google_project = MagicMock(spec=GoogleProject)
        google_project.project_id = "test_project_id"
        domain = "example.com"

        expected_dir = "test_provider_id/group/example.com/templates"

        assert get_response_dir(exe_message, google_project, domain).endswith(
            expected_dir
        )

    def test_get_group_dir(self):
        base_dir = "test_base_dir"
        domain = "example.com"
        expected_path = os.path.join(
            "test_base_dir",
            "resources",
            "google_workspace",
            "group",
            "example.com",
        )
        self.assertEqual(get_group_dir(base_dir, domain), expected_path)

    def test_get_templated_resource_file_path(self):
        resource_dir = "test_resource_dir"
        resource_email = "test_group@example.com"
        expected_path = os.path.join("test_resource_dir", "test_group.yaml")
        self.assertEqual(
            get_templated_resource_file_path(resource_dir, resource_email),
            expected_path,
        )

    @patch(
        "iambic.plugins.v0_1_0.google_workspace.group.template_generation.list_groups"
    )
    @patch(
        "iambic.plugins.v0_1_0.google_workspace.group.template_generation.GoogleWorkspaceGroupTemplate.write"
    )
    async def test_generate_domain_group_resource_files(
        self, mock_group_write, mock_list_groups
    ):
        exe_message = ExecutionMessage(
            provider_id="test_provider_id",
            execution_id="test_execution_id",
            command="import",
        )
        project = MagicMock(spec=GoogleProject)
        project.project_id = "test_project_id"
        domain = "example.com"

        group = GoogleWorkspaceGroupTemplate(
            file_path="unset",
            properties={
                "name": "Group 1",
                "email": "group1@example.com",
                "description": "Group 1 description",
                "members": [],
                "domain": "example.com",
            },
        )
        mock_list_groups.return_value = [group]

        await generate_domain_group_resource_files(exe_message, project, domain)

        mock_list_groups.assert_called_once_with(domain, project)
        mock_group_write.assert_called_once()

    async def test_update_or_create_group_template(self):
        existing_template_map = {}
        group_dir = "test_group_dir"
        discovered_group_template = GoogleWorkspaceGroupTemplate(
            file_path="unset",
            properties={
                "name": "Group 1",
                "email": "group1@example.com",
                "description": "Group 1 description",
                "members": [],
                "domain": "example.com",
            },
        )

        with patch(
            "iambic.plugins.v0_1_0.google_workspace.group.template_generation.common_create_or_update_template"
        ) as mock_common_create_or_update_template:
            mock_common_create_or_update_template.return_value = (
                discovered_group_template
            )

            result = await update_or_create_group_template(
                discovered_group_template, existing_template_map, group_dir
            )

            mock_common_create_or_update_template.assert_called_once_with(
                discovered_group_template.file_path,
                existing_template_map,
                discovered_group_template.resource_id,
                GoogleWorkspaceGroupTemplate,
                {},
                discovered_group_template.properties,
                [],
            )
            self.assertEqual(result, discovered_group_template)

    @patch(
        "iambic.plugins.v0_1_0.google_workspace.group.template_generation.generate_domain_group_resource_files"
    )
    async def test_collect_project_groups(
        self, mock_generate_domain_group_resource_files
    ):
        exe_message = ExecutionMessage(
            provider_id="test_provider_id",
            execution_id="test_execution_id",
            command="import",
        )
        exe_message.provider_id = "test_project_id"
        project = MagicMock(spec=GoogleProject)
        project.project_id = "test_project_id"
        subject = GoogleSubject(
            domain="example.com",
            service_account="google-groups-test@example-ga.iam.gserviceaccount.com",
        )
        config = GoogleWorkspaceConfig(
            workspaces=[
                GoogleProject(
                    project_id="example-ga",
                    project_name=None,
                    subjects=[subject],
                    type="service_account",
                    private_key_id="12345",
                    private_key=SecretStr("**********"),
                    client_email="google-groups-test@example-ga.iam.gserviceaccount.com",
                    client_id="12345",
                    auth_uri="https://accounts.google.com/o/oauth2/auth",
                    token_uri="https://oauth2.googleapis.com/token",
                    auth_provider_x509_cert_url="https://www.googleapis.com/oauth2/v1/certs",
                    client_x509_cert_url="https://www.googleapis.com/robot/v1/metadata/x509/google-groups-test%40example-ga.iam.gserviceaccount.com",
                    variables=[],
                    iambic_managed=IambicManaged.UNDEFINED,
                )
            ]
        )
        with self.assertRaises(Exception):
            await collect_project_groups(exe_message, config)

        mock_generate_domain_group_resource_files.assert_not_called()
        mock_generate_domain_group_resource_files.reset_mock()
        project.project_id = "example-ga"
        exe_message.provider_id = "example-ga"
        await collect_project_groups(exe_message, config)
        mock_generate_domain_group_resource_files.assert_called_once()



from __future__ import annotations

import itertools
import os
from collections import defaultdict
from typing import TYPE_CHECKING

import aiofiles

from iambic.core import noq_json as json
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
)
from iambic.core.utils import NoqSemaphore, normalize_dict_keys, resource_file_upsert
from iambic.plugins.v0_1_0.aws.event_bridge.models import RoleMessageDetails
from iambic.plugins.v0_1_0.aws.iam.policy.models import AssumeRolePolicyDocument
from iambic.plugins.v0_1_0.aws.iam.role.models import (
    AWS_IAM_ROLE_TEMPLATE_TYPE,
    AwsIamRoleTemplate,
    RoleProperties,
)
from iambic.plugins.v0_1_0.aws.iam.role.utils import (
    get_role_across_accounts,
    get_role_inline_policies,
    get_role_managed_policies,
    list_role_tags,
    list_roles,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.template_generation import (
    base_group_str_attribute,
    group_dict_attribute,
    group_int_or_str_attribute,
)
from iambic.plugins.v0_1_0.aws.utils import (
    calculate_import_preference,
    get_aws_account_map,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

RESOURCE_DIR = ["iam", "role"]


def get_response_dir(exe_message: ExecutionMessage, aws_account: AWSAccount) -> str:
    if exe_message.provider_id:
        return exe_message.get_directory(*RESOURCE_DIR)
    else:
        return exe_message.get_directory(aws_account.account_id, *RESOURCE_DIR)


def get_template_dir(base_dir: str) -> str:
    return str(os.path.join(base_dir, "resources", "aws", *RESOURCE_DIR))


def get_templated_role_file_path(
    role_dir: str,
    role_name: str,
    included_accounts: list[str],
) -> str:
    if included_accounts is not None and len(included_accounts) > 1:
        separator = "multi_account"
    elif included_accounts == ["*"] or included_accounts is None:
        separator = "all_accounts"
    else:
        separator = included_accounts[0]

    file_name = (
        role_name.replace(" ", "")
        .replace("{{var.", "")
        .replace("{{", "")
        .replace("}}_", "_")
        .replace("}}", "_")
        .replace(".", "_")
        .lower()
    )
    return str(os.path.join(role_dir, separator, f"{file_name}.yaml"))


async def generate_account_role_resource_files(
    exe_message: ExecutionMessage,
    aws_account: AWSAccount,
) -> dict:
    account_resource_dir = get_response_dir(exe_message, aws_account)
    role_resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []

    response = dict(account_id=aws_account.account_id, roles=[])
    iam_client = await aws_account.get_boto3_client("iam")
    account_roles = await list_roles(iam_client)

    log.debug(
        "Retrieved AWS IAM Roles.",
        account_id=aws_account.account_id,
        account_name=aws_account.account_name,
        role_count=len(account_roles),
    )

    for account_role in account_roles:
        role_path = os.path.join(
            account_resource_dir, f'{account_role["RoleName"]}.json'
        )
        response["roles"].append(
            {
                "path": role_path,
                "name": account_role["RoleName"],
                "account_id": aws_account.account_id,
            }
        )
        messages.append(
            dict(file_path=role_path, content_as_dict=account_role, replace_file=True)
        )

    await role_resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM Roles.",
        account_id=aws_account.account_id,
        role_count=len(account_roles),
    )

    return response


async def generate_role_resource_file_for_all_accounts(
    exe_message: ExecutionMessage,
    aws_accounts: list[AWSAccount],
    role_name: str,
) -> list:
    account_resource_dir_map = {
        aws_account.account_id: get_response_dir(exe_message, aws_account)
        for aws_account in aws_accounts
    }
    role_resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []
    response = []

    role_across_accounts = await get_role_across_accounts(
        aws_accounts, role_name, False
    )
    role_across_accounts = {k: v for k, v in role_across_accounts.items() if v}

    log.debug(
        "Retrieved AWS IAM Role for all accounts.",
        role_name=role_name,
        total_accounts=len(role_across_accounts),
    )

    for account_id, account_role in role_across_accounts.items():
        role_path = os.path.join(
            account_resource_dir_map[account_id],
            f'{account_role["RoleName"]}.json',
        )
        response.append(
            {
                "path": role_path,
                "name": account_role["RoleName"],
                "account_id": account_id,
            }
        )
        messages.append(
            dict(file_path=role_path, content_as_dict=account_role, replace_file=True)
        )

    await role_resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM Role for all accounts.",
        role_name=role_name,
        total_accounts=len(role_across_accounts),
    )

    return response


async def set_role_resource_tags(
    role_name: str, role_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    role_tags = await list_role_tags(role_name, iam_client)
    await resource_file_upsert(role_resource_path, {"Tags": role_tags}, False)


async def set_role_resource_inline_policies(
    role_name: str, role_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    role_inline_policies = await get_role_inline_policies(role_name, iam_client)
    for k in role_inline_policies.keys():
        role_inline_policies[k]["policy_name"] = k

    role_inline_policies = list(role_inline_policies.values())
    await resource_file_upsert(
        role_resource_path, {"InlinePolicies": role_inline_policies}, False
    )


async def set_role_resource_managed_policies(
    role_name: str, role_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    role_managed_policies = await get_role_managed_policies(role_name, iam_client)
    await resource_file_upsert(
        role_resource_path, {"ManagedPolicies": role_managed_policies}, False
    )


async def _account_id_to_role_map(role_refs):
    account_id_to_role_map = {}
    for role_ref in role_refs:
        async with aiofiles.open(role_ref["path"], mode="r") as f:
            content_dict = json.loads(await f.read())

            # handle strange unstable response with deleted principals
            assume_role_policy_document = content_dict.get(
                "AssumeRolePolicyDocument", None
            )
            if assume_role_policy_document:
                policy_document = AssumeRolePolicyDocument.parse_obj(
                    normalize_dict_keys(assume_role_policy_document)
                )
                content_dict["AssumeRolePolicyDocument"] = json.loads(
                    policy_document.json(
                        exclude_unset=True, exclude_defaults=True, exclude_none=True
                    )
                )

            account_id_to_role_map[role_ref["account_id"]] = normalize_dict_keys(
                content_dict
            )
    return account_id_to_role_map


async def create_templated_role(  # noqa: C901
    aws_account_map: dict[str, AWSAccount],
    role_name: str,
    role_refs: list[dict],
    role_dir: str,
    existing_template_map: dict,
    config: AWSConfig,
) -> AwsIamRoleTemplate:
    account_id_to_role_map = await _account_id_to_role_map(role_refs)
    num_of_accounts = len(role_refs)

    min_accounts_required_for_wildcard_included_accounts = (
        config.min_accounts_required_for_wildcard_included_accounts
    )

    # calculate preference based on existing template
    prefer_templatized = calculate_import_preference(
        existing_template_map.get(role_name)
    )

    # Generate the params used for attribute creation
    role_template_params = {"identifier": role_name}
    role_template_properties = {"role_name": role_name}
    path_resources = list()
    description_resources = list()
    managed_policy_resources = list()
    assume_role_policy_document_resources = list()
    inline_policy_document_resources = list()
    permissions_boundary_resources = list()
    tag_resources = list()
    max_session_duration_resources = dict()
    for account_id, role_dict in account_id_to_role_map.items():
        max_session_duration_resources[account_id] = role_dict["max_session_duration"]
        path_resources.append(
            {
                "account_id": account_id,
                "resources": [{"resource_val": role_dict["path"]}],
            }
        )
        assume_role_policy_document_resources.append(
            {
                "account_id": account_id,
                "resources": [
                    {"resource_val": role_dict["assume_role_policy_document"]}
                ],
            }
        )

        if managed_policies := role_dict.get("managed_policies"):
            managed_policy_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": mp} for mp in managed_policies],
                }
            )

        if permissions_boundary := role_dict.get("permissions_boundary"):
            permissions_boundary_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": permissions_boundary}],
                }
            )

        if inline_policies := role_dict.get("inline_policies"):
            # Normalize the inline policy statements to be a list
            for inline_policy in inline_policies:
                if isinstance(inline_policy.get("statement"), dict):
                    inline_policy["statement"] = [inline_policy["statement"]]
            inline_policy_document_resources.append(
                {
                    "account_id": account_id,
                    "resources": [
                        {"resource_val": inline_policy}
                        for inline_policy in inline_policies
                    ],
                }
            )

        if tags := role_dict.get("tags"):
            tag_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": tag} for tag in tags],
                }
            )

        if description := role_dict.get("description"):
            description_resources.append(
                {"account_id": account_id, "resources": [{"resource_val": description}]}
            )

    if (
        len(role_refs) != len(aws_account_map)
        or len(aws_account_map) <= min_accounts_required_for_wildcard_included_accounts
    ):
        role_template_params["included_accounts"] = [
            aws_account_map[role_ref["account_id"]].account_name
            for role_ref in role_refs
        ]
    else:
        role_template_params["included_accounts"] = ["*"]

    path = await group_int_or_str_attribute(
        aws_account_map, num_of_accounts, path_resources, "path"
    )
    if path != "/":
        role_template_properties["path"] = path

    max_session_duration = await group_int_or_str_attribute(
        aws_account_map,
        num_of_accounts,
        max_session_duration_resources,
        "max_session_duration",
    )
    if max_session_duration != 3600:
        role_template_properties["max_session_duration"] = max_session_duration

    if assume_role_policy_document_resources:
        role_template_properties[
            "assume_role_policy_document"
        ] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            assume_role_policy_document_resources,
            prefer_templatized=prefer_templatized,
        )

    if permissions_boundary_resources:
        role_template_properties["permissions_boundary"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            permissions_boundary_resources,
            prefer_templatized=prefer_templatized,
        )

    if description_resources:
        role_template_properties["description"] = await group_int_or_str_attribute(
            aws_account_map, num_of_accounts, description_resources, "description"
        )

    if managed_policy_resources:
        role_template_properties["managed_policies"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            managed_policy_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if inline_policy_document_resources:
        role_template_properties["inline_policies"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            inline_policy_document_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if tag_resources:
        tags = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            tag_resources,
            True,
            prefer_templatized=prefer_templatized,
        )
        if isinstance(tags, dict):
            tags = [tags]

        role_template_properties["tags"] = tags

    file_path = get_templated_role_file_path(
        role_dir,
        role_name,
        role_template_params.get("included_accounts"),
    )
    return create_or_update_template(
        file_path,
        existing_template_map,
        role_name,
        AwsIamRoleTemplate,
        role_template_params,
        RoleProperties(**role_template_properties),
        list(aws_account_map.values()),
    )


async def collect_aws_roles(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    iam_template_map: dict,
    detect_messages: list[RoleMessageDetails] = None,
):
    aws_account_map = await get_aws_account_map(config)
    if exe_message.provider_id:
        aws_account_map = {
            exe_message.provider_id: aws_account_map[exe_message.provider_id]
        }

    existing_template_map = iam_template_map.get(AWS_IAM_ROLE_TEMPLATE_TYPE, {})
    set_role_resource_inline_policies_semaphore = NoqSemaphore(
        set_role_resource_inline_policies, 20
    )
    set_role_resource_managed_policies_semaphore = NoqSemaphore(
        set_role_resource_managed_policies, 30
    )
    set_role_resource_tags_semaphore = NoqSemaphore(set_role_resource_tags, 45)

    log.info(
        "Generating AWS role templates. Beginning to retrieve AWS IAM Roles.",
        accounts=list(aws_account_map.keys()),
    )

    if detect_messages:
        aws_accounts = list(aws_account_map.values())
        generate_role_resource_file_for_all_accounts_semaphore = NoqSemaphore(
            generate_role_resource_file_for_all_accounts, 45
        )
        tasks = [
            {
                "exe_message": exe_message,
                "aws_accounts": aws_accounts,
                "role_name": role.role_name,
            }
            for role in detect_messages
            if not role.delete
        ]

        # Remove deleted or mark templates for update
        deleted_roles = [role for role in detect_messages if role.delete]
        if deleted_roles:
            for role in deleted_roles:
                role_account = aws_account_map.get(role.account_id)
                if existing_template := existing_template_map.get(role.role_name):
                    if len(existing_template.included_accounts) == 1 and (
                        existing_template.included_accounts[0]
                        == role_account.account_name
                        or existing_template.included_accounts[0]
                        == role_account.account_id
                    ):
                        # It's the only account for the template so delete it
                        existing_template.delete()
                    else:
                        # There are other accounts for the template so re-eval the template
                        tasks.append(
                            {
                                "exe_message": exe_message,
                                "aws_accounts": aws_accounts,
                                "role_name": existing_template.properties.role_name,
                            }
                        )

        account_role_list = (
            await generate_role_resource_file_for_all_accounts_semaphore.process(tasks)
        )
        account_role_list = list(itertools.chain.from_iterable(account_role_list))
        account_role_map = defaultdict(list)
        for account_role in account_role_list:
            account_role_map[account_role["account_id"]].append(account_role)
        account_roles = [
            dict(account_id=account_id, roles=roles)
            for account_id, roles in account_role_map.items()
        ]
    elif exe_message.provider_id:
        aws_account = aws_account_map[exe_message.provider_id]
        account_roles = [
            (await generate_account_role_resource_files(exe_message, aws_account))
        ]
    else:
        generate_account_role_resource_files_semaphore = NoqSemaphore(
            generate_account_role_resource_files, 5
        )
        account_roles = await generate_account_role_resource_files_semaphore.process(
            [
                {"aws_account": aws_account, "exe_message": exe_message}
                for aws_account in aws_account_map.values()
            ]
        )

    messages = []
    # Upsert roles
    for account_role in account_roles:
        for role in account_role["roles"]:
            messages.append(
                {
                    "role_name": role["name"],
                    "role_resource_path": role["path"],
                    "aws_account": aws_account_map[account_role["account_id"]],
                }
            )

    log.info(
        "Setting inline policies in role templates",
        accounts=list(aws_account_map.keys()),
    )
    await set_role_resource_inline_policies_semaphore.process(messages)
    log.info(
        "Setting managed policies in role templates",
        accounts=list(aws_account_map.keys()),
    )
    await set_role_resource_managed_policies_semaphore.process(messages)
    log.info("Setting tags in role templates", accounts=list(aws_account_map.keys()))
    await set_role_resource_tags_semaphore.process(messages)
    log.info("Finished retrieving role details", accounts=list(aws_account_map.keys()))

    account_role_output = json.dumps(account_roles)
    with open(
        exe_message.get_file_path(*RESOURCE_DIR, file_name_and_extension="output.json"),
        "w",
    ) as f:
        f.write(account_role_output)


async def generate_aws_role_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    iam_template_map: dict,
    detect_messages: list[RoleMessageDetails] = None,
):
    role_dir = get_template_dir(base_output_dir)
    aws_account_map = await get_aws_account_map(config)
    existing_template_map = iam_template_map.get(AWS_IAM_ROLE_TEMPLATE_TYPE, {})
    account_roles = await exe_message.get_sub_exe_files(
        *RESOURCE_DIR, file_name_and_extension="output.json", flatten_results=True
    )

    if detect_messages:
        detect_messages = [
            msg for msg in detect_messages if isinstance(msg, RoleMessageDetails)
        ]
        if not detect_messages:
            return

    log.info("Grouping roles")
    # Move everything to required structure
    for account_role_elem in range(len(account_roles)):
        for role_elem in range(len(account_roles[account_role_elem]["roles"])):
            role_name = account_roles[account_role_elem]["roles"][role_elem].pop("name")
            account_roles[account_role_elem]["roles"][role_elem][
                "resource_val"
            ] = role_name

        account_roles[account_role_elem]["resources"] = account_roles[
            account_role_elem
        ].pop("roles", [])

    grouped_role_map = await base_group_str_attribute(aws_account_map, account_roles)

    log.info("Writing templated roles")
    all_resource_ids = set()
    for role_name, role_refs in grouped_role_map.items():
        resource_template = await create_templated_role(
            aws_account_map,
            role_name,
            role_refs,
            role_dir,
            existing_template_map,
            config,
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    if not detect_messages:
        # NEVER call this if messages are passed in because all_resource_ids will only contain those resources
        delete_orphaned_templates(
            list(existing_template_map.values()), all_resource_ids
        )

    log.info("Finished templated role generation")



from __future__ import annotations

import asyncio
import json
import typing
from enum import Enum
from itertools import chain
from typing import TYPE_CHECKING, Any, List, Optional, Union

from aiohttp import ClientResponseError
from pydantic import Field

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.core.utils import normalize_dict_keys
from iambic.plugins.v0_1_0.azure_ad.models import AzureADOrganization, AzureADTemplate

if TYPE_CHECKING:
    MappingIntStrAny = typing.Mapping[int | str, any]
    AbstractSetIntStr = typing.AbstractSet[int | str]

AZURE_AD_GROUP_TEMPLATE_TYPE = "NOQ::AzureAD::Group"


class MemberDataType(Enum):
    USER = "user"
    GROUP = "group"


class Member(BaseModel, ExpiryModel):
    id: Optional[str] = None
    name: str
    data_type: MemberDataType
    """TODO: validate name

    https://learn.microsoft.com/en-us/azure/active-directory/authentication/concept-sspr-policy#userprincipalname-policies-that-apply-to-all-user-accounts

    The total length must not exceed 113 characters
    There can be up to 64 characters before the "@" symbol
    There can be up to 48 characters after the "@" symbol
    """

    @property
    def resource_type(self) -> str:
        return "azure_ad:user"

    @property
    def resource_id(self) -> str:
        return f"{self.name} - {self.id}"

    def dict(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":  # noqa
        if not exclude:
            exclude = {"metadata_commented_dict", "deleted"}
        else:
            exclude.add("metadata_commented_dict")
            exclude.add("deleted")

        return json.loads(
            super().json(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )
        )


class GroupTemplateProperties(ExpiryModel, BaseModel):
    name: str = Field(..., description="Name of the group", max_length=256)
    mail_nickname: Optional[str] = Field(
        None,
        description="Mail nickname of the group",
        regex=r"^[!#$%&'*+-./0-9=?A-Z^_`a-z{|}~]{1,64}$",
    )
    group_id: Optional[str] = Field(
        None,
        description="Unique Group ID for the group. Usually it's {idp-name}-{name}",
    )
    description: Optional[str] = Field("", description="Description of the group")
    group_types: Optional[list[str]] = Field(
        [], description="Specifies the group type and its membership."
    )
    mail: Optional[str] = Field(description="Email address of the group")
    mail_enabled: Optional[bool] = False
    security_enabled: Optional[bool] = True
    extra: Optional[Any] = Field(None, description="Extra attributes to store")
    is_assignable_to_role: Optional[bool] = Field(
        description="Indicates whether this group can be assigned to an Azure Active Directory role or not."
    )
    membership_rule: Optional[str] = Field(
        description="The rule that determines members for this group if the group is a dynamic group."
    )
    members: Optional[List[Member]] = Field(
        [], description="A list of users in the group"
    )

    classification: Optional[str] = Field(
        None,
        description="Classification of the group",
        exclude=True,
    )
    created_date_time: Optional[str] = Field(
        None,
        description="Date and time when the group was created",
        exclude=True,
    )
    creation_options: Optional[List[str]] = Field(
        None,
        description="Specifies the group type and its membership",
        exclude=True,
    )
    deleted_date_time: Optional[str] = Field(
        None,
        description="Date and time when the group was deleted",
        exclude=True,
    )
    expiration_date_time: Optional[str] = Field(
        None,
        description="Date and time when the group expires",
        exclude=True,
    )
    membership_rule_processing_state: Optional[str] = Field(
        None,
        description="Indicates whether the group is a dynamic group or not",
        exclude=True,
    )
    preferred_data_location: Optional[str] = Field(
        None,
        description="Preferred data location of the group",
        exclude=True,
    )
    preferred_language: Optional[str] = Field(
        None,
        description="Preferred language of the group",
        exclude=True,
    )
    proxy_addresses: Optional[List[str]] = Field(
        None,
        description="Proxy addresses of the group",
        exclude=True,
    )
    renewed_date_time: Optional[str] = Field(
        None,
        description="Date and time when the group was renewed",
        exclude=True,
    )
    resource_behavior_options: Optional[List[str]] = Field(
        None,
        description="Resource behavior options of the group",
        exclude=True,
    )
    resource_provisioning_options: Optional[List[str]] = Field(
        None,
        description="Resource provisioning options of the group",
        exclude=True,
    )
    security_identifier: Optional[str] = Field(
        None,
        description="Security identifier of the group",
        exclude=True,
    )
    theme: Optional[str] = Field(
        None,
        description="Theme of the group",
        exclude=True,
    )
    visibility: Optional[str] = Field(
        None,
        description="Visibility of the group",
        exclude=True,
    )
    extra: Any = Field(None, description=("Extra attributes to store"), exclude=True)

    def __init__(self, **data):
        if "mail_nickname" not in data:
            data["mail_nickname"] = data.get("name")
        super().__init__(**data)

    @property
    def resource_type(self) -> str:
        return "azure_ad:group"

    @property
    def resource_id(self) -> str:
        return self.mail or self.name

    @classmethod
    def from_azure_response(cls, azure_response: dict):
        azure_response = normalize_dict_keys(azure_response)
        # Filter unwanted keys
        group_id = azure_response.pop("id")
        name = azure_response.pop("display_name")
        azure_response = {
            x: y for x, y in azure_response.items() if not x.startswith("@odata")
        }
        azure_response = {
            x: y for x, y in azure_response.items() if not x.startswith("on_premises")
        }
        return cls(
            group_id=group_id,
            name=name,
            members=[],
            **azure_response,
        )


class AzureActiveDirectoryGroupTemplate(ExpiryModel, AzureADTemplate):
    template_type = AZURE_AD_GROUP_TEMPLATE_TYPE
    owner: Optional[str] = Field(None, description="Owner of the group")
    properties: GroupTemplateProperties = Field(
        ..., description="Properties for the Azure AD Group"
    )

    @property
    def resource_type(self) -> str:
        return "azure_ad:group"

    def apply_resource_dict(self, azure_ad_organization: AzureADOrganization):
        return {
            "name": self.properties.name,
            "description": self.properties.description,
            "members": self.properties.members,
        }

    async def _apply_to_account(
        self, azure_ad_organization: AzureADOrganization
    ) -> AccountChangeDetails:
        from iambic.plugins.v0_1_0.azure_ad.group.utils import (
            create_group,
            delete_group,
            get_group,
            resolve_member_ids,
            update_group_attributes,
            update_group_members,
        )

        change_details = AccountChangeDetails(
            account=self.idp_name,
            resource_id=self.resource_id,
            new_value=self.properties.dict(
                exclude={"metadata_commented_dict", "deleted"}, exclude_none=True
            ),
            proposed_changes=[],
        )
        cloud_group = None

        log_params = dict(
            resource_type=self.resource_type,
            resource_id=self.resource_id,
            organization=str(self.idp_name),
        )

        if self.properties.group_id and not self.deleted:
            try:
                cloud_group: Optional[GroupTemplateProperties] = await get_group(
                    azure_ad_organization,
                    self.properties.group_id,
                    self.properties.name,
                )
                change_details.current_value = cloud_group
            except ClientResponseError as err:
                if err.status == 404:
                    err = f"Group not found in Azure AD where id={self.properties.group_id}"
                    log.exception(
                        "Invalid group_id provided. Group not found in Azure AD.",
                        **log_params,
                    )
                change_details.extend_changes(
                    [
                        ProposedChange(
                            change_type=ProposedChangeType.UPDATE,
                            resource_id=self.resource_id,
                            resource_type=self.resource_type,
                            exceptions_seen=[str(err)],
                        )
                    ]
                )
                return change_details

        group_exists = bool(cloud_group)
        tasks = []

        await self.remove_expired_resources()

        self.properties.members = await resolve_member_ids(
            azure_ad_organization, self.properties.members
        )

        if not group_exists and not self.deleted:
            log_str = "New resource found in code."
            change_details.extend_changes(
                [
                    ProposedChange(
                        change_type=ProposedChangeType.CREATE,
                        resource_id=self.resource_id,
                        resource_type=self.resource_type,
                    )
                ]
            )
            if not ctx.execute:
                log.info(log_str, **log_params)
                # Exit now because apply functions won't work if resource doesn't exist
                return change_details

            log_str = f"{log_str} Creating resource..."
            log.info(log_str, **log_params)

            try:
                cloud_group: GroupTemplateProperties = await create_group(
                    azure_ad_organization=azure_ad_organization,
                    group_name=self.properties.name,
                    description=self.properties.description,
                    mail_enabled=self.properties.mail_enabled,
                    mail_nickname=self.properties.mail_nickname,
                    security_enabled=self.properties.security_enabled,
                    group_types=self.properties.group_types,
                )
                self.properties.group_id = cloud_group.group_id
                self.write()
            except ClientResponseError as err:
                log.exception(
                    "Failed to create user in Azure AD",
                    **log_params,
                )
                proposed_change = change_details.proposed_changes.pop(-1)
                proposed_change.exceptions_seen.append(str(err))
                change_details.extend_changes([proposed_change])
                return change_details

        if self.deleted:
            change_details.extend_changes(
                await delete_group(
                    azure_ad_organization,
                    self.properties,
                    log_params,
                )
            )
        else:
            tasks.extend(
                [
                    update_group_attributes(
                        azure_ad_organization,
                        self.properties,
                        cloud_group,
                        log_params,
                    ),
                    update_group_members(
                        azure_ad_organization,
                        cloud_group,
                        [
                            member
                            for member in self.properties.members
                            if not member.deleted
                        ],
                        log_params,
                    ),
                ]
            )

            changes_made = await asyncio.gather(*tasks, return_exceptions=True)
            if any(changes_made):
                change_details.extend_changes(list(chain.from_iterable(changes_made)))

        if ctx.execute and not change_details.exceptions_seen:
            log.debug(
                "Successfully finished execution for resource",
                changes_made=bool(change_details.proposed_changes),
                **log_params,
            )
        elif change_details.exceptions_seen:
            cmd_verb = "apply" if ctx.execute else "scan for"
            log.error(
                f"Failed to successfully {cmd_verb} resource changes",
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift for resource",
                requires_changes=bool(change_details.proposed_changes),
                **log_params,
            )

        return change_details



from __future__ import annotations

import asyncio
import json

import okta.models as models
from okta.errors.okta_api_error import OktaAPIError

from iambic.core.exceptions import RateLimitException
from iambic.plugins.v0_1_0.okta.exceptions import UserProfileNotUpdatableYet


async def generate_user_profile(user: models.User):
    """
    Generates a key-value pair of user profile attributes that aren't None.
    This is useful to keep user templates small.
    """
    return {k: v for (k, v) in user.profile.__dict__.items() if v is not None}


async def handle_okta_fn(fn, *args, **kwargs):
    try:
        res = await fn(*args, **kwargs)
    except asyncio.exceptions.TimeoutError:
        raise asyncio.exceptions.TimeoutError

    err = res[-1]
    if err:
        if isinstance(err, Exception):
            raise err
        # Handle the case where Okta SDK returns appropriately scoped OktaAPIError
        if isinstance(err, OktaAPIError):
            if err.error_code == "E0000047":
                raise RateLimitException(err.error_summary)
            if err.error_code == "E0000112":
                raise UserProfileNotUpdatableYet(
                    "Unable to update profile, user is not fully provisioned"
                )
            return res

        # Handle the case where Okta SDK returns JSON
        try:
            err_j = json.loads(err)
            if err_j.get("errorCode") == "E0000047":
                raise RateLimitException(err)
        except TypeError:
            pass
    return res



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
    get_existing_template_map,
)
from iambic.plugins.v0_1_0.okta.group.utils import list_all_users
from iambic.plugins.v0_1_0.okta.user.models import (
    OKTA_USER_TEMPLATE_TYPE,
    OktaUserTemplate,
    UserProperties,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaConfig


def get_resource_dir_args() -> list:
    return ["user"]


def get_response_dir(exe_message: ExecutionMessage) -> str:
    return exe_message.get_directory(*get_resource_dir_args(), "templates")


async def update_or_create_user_template(
    discovered_template: OktaUserTemplate, existing_template_map: dict
) -> OktaUserTemplate:
    return create_or_update_template(
        discovered_template.file_path,
        existing_template_map,
        discovered_template.resource_id,
        OktaUserTemplate,
        {"idp_name": discovered_template.idp_name},
        discovered_template.properties,
        [],
    )


async def collect_org_users(exe_message: ExecutionMessage, config: OktaConfig):
    assert exe_message.provider_id
    base_path = get_response_dir(exe_message)
    okta_organization = config.get_organization(exe_message.provider_id)
    log.info("Beginning to retrieve Okta users.", organization=exe_message.provider_id)

    users = await list_all_users(okta_organization)
    for user in users:
        okta_user = OktaUserTemplate(
            file_path="unset",
            idp_name=user.idp_name,
            properties=UserProperties(
                username=user.username,
                user_id=user.user_id,
                status=user.status.value,
                profile=user.profile,
            ),
        )
        okta_user.file_path = os.path.join(base_path, f"{user.username}.yaml")
        okta_user.write()

    log.info(
        "Finished retrieving Okta users.",
        okta_org=exe_message.provider_id,
        user_count=len(users),
    )


async def generate_user_templates(
    exe_message: ExecutionMessage, output_dir: str, detect_messages: list = None
):
    """List all users in the domain, along with members and settings"""
    base_path = os.path.expanduser(output_dir)
    existing_template_map = await get_existing_template_map(
        base_path, OKTA_USER_TEMPLATE_TYPE
    )
    all_resource_ids = set()

    log.info("Updating and creating Okta user templates.")

    users = await exe_message.get_sub_exe_files(
        *get_resource_dir_args(), "templates", file_name_and_extension="**.yaml"
    )
    # Update or create templates
    for user in users:
        user = OktaUserTemplate(file_path="unset", **user)
        user.set_default_file_path(output_dir)
        resource_template = await update_or_create_user_template(
            user, existing_template_map
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    # Delete templates that no longer exist
    delete_orphaned_templates(list(existing_template_map.values()), all_resource_ids)

    log.info("Finish updating and creating Okta user templates.")



from __future__ import annotations

from iambic.core.context import ctx
from iambic.core.models import TemplateChangeDetails
from iambic.request_handler.git_apply import apply_git_changes


async def plan_git_changes(
    config_path: str, repo_dir: str
) -> list[TemplateChangeDetails]:
    """Retrieves files added/updated/or removed when comparing the current branch to master

    Works by taking the diff and adding implied changes to the templates that were modified.
    These are the changes and this function detects:
        Deleting a file
        Removing 1 or more aws_accounts from included_accounts
        Adding 1 or more aws_accounts to excluded_accounts

    :param config_path:
    :param repo_dir:
    :return:
    """
    ctx_eval_only_original_value = ctx.eval_only
    ctx.eval_only = True
    try:
        changes = await apply_git_changes(config_path, repo_dir)
    finally:
        ctx.eval_only = ctx_eval_only_original_value
        return changes



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, List

import okta.models as models

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import GlobalRetryController
from iambic.plugins.v0_1_0.okta.group.utils import get_group
from iambic.plugins.v0_1_0.okta.models import App, Assignment, Group
from iambic.plugins.v0_1_0.okta.utils import handle_okta_fn

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaOrganization


async def list_app_user_assignments(
    okta_organization: OktaOrganization, app: App
) -> dict:
    client = await okta_organization.get_okta_client()
    async with GlobalRetryController(
        fn_identifier="okta.list_application_users"
    ) as retry_controller:
        fn = functools.partial(client.list_application_users, app.id)
        app_user_list, _, err = await retry_controller(handle_okta_fn, fn)
    if err:
        log.error("Error encountered when listing app users", error=str(err))
        raise Exception(f"Error listing app users: {str(err)}")

    user_assignments = []
    if app_user_list:
        for user in app_user_list:
            if user.scope == "GROUP":
                continue
            async with GlobalRetryController(
                fn_identifier="okta.get_user"
            ) as retry_controller:
                fn = functools.partial(client.get_user, user.id)
                user_okta, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                if isinstance(err, asyncio.exceptions.TimeoutError):
                    raise err
                log.error("Error encountered when getting user", error=str(err))
                raise Exception("Error encountered when getting user")
            user_assignments.append(user_okta.profile.login)

    return {
        "app_id": app.id,
        "user_assignments": user_assignments,
    }


async def get_app(okta_organization: OktaOrganization, app_id: str) -> App:
    client = await okta_organization.get_okta_client()
    async with GlobalRetryController(
        fn_identifier="okta.get_application"
    ) as retry_controller:
        fn = functools.partial(client.get_application, app_id)
        app_raw, _, err = await retry_controller(handle_okta_fn, fn)
    if err:
        if isinstance(err, asyncio.exceptions.TimeoutError):
            raise err
        log.error("Error encountered when getting app", error=str(err))
        raise Exception(f"Error encountered when getting app: {str(err)}")

    app = App(
        id=app_raw.id,
        idp_name=okta_organization.idp_name,
        name=app_raw.label,
        attributes=dict(),
        extra=dict(
            okta_app_id=app_raw.id,
            created=app_raw.created,
        ),
    )

    user_assignments = await list_app_user_assignments(okta_organization, app)
    user_assignments = user_assignments.get("user_assignments", [])
    group_assignments = await list_app_group_assignments(okta_organization, app)
    group_assignments = group_assignments.get("group_assignments", [])

    for assignment in user_assignments:
        app.assignments.append(Assignment(user=assignment))
    for assignment in group_assignments:
        app.assignments.append(Assignment(group=assignment))
    return app


async def list_app_group_assignments(
    okta_organization: OktaOrganization, app: App
) -> dict:
    client = await okta_organization.get_okta_client()
    async with GlobalRetryController(
        fn_identifier="okta.list_application_group_assignments"
    ) as retry_controller:
        fn = functools.partial(client.list_application_group_assignments, app.id)
        app_group_assignments, _, err = await retry_controller(handle_okta_fn, fn)

    if err:
        log.error(
            "Error encountered when listing app group assignments", error=str(err)
        )
        raise Exception(
            f"Error encountered when listing app group assignments: {str(err)}"
        )
    groups_assignments = []
    for assignment in app_group_assignments:
        async with GlobalRetryController(
            fn_identifier="okta.get_group"
        ) as retry_controller:
            fn = functools.partial(client.get_group, assignment.id)
            group, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            log.error(
                "Error encountered when getting group",
                group_id=assignment.id,
                error=str(err),
            )
            continue
        groups_assignments.append(group.profile.name)
    return {
        "app_id": app.id,
        "group_assignments": groups_assignments,
    }


async def list_all_apps(okta_organization: OktaOrganization) -> List[App]:
    """
    List all apps in Okta.

    Args:
    - okta_organization: An instance of the OktaOrganization class, which provides access to the Okta API.

    Returns:
    - A list of `App` instances, representing the apps in Okta.
    """

    client = await okta_organization.get_okta_client()
    log.info("Listing apps", provder="Okta", organization=okta_organization.idp_name)
    async with GlobalRetryController(
        fn_identifier="okta.list_applications"
    ) as retry_controller:
        fn = functools.partial(client.list_applications)
        raw_apps, resp, err = await retry_controller(handle_okta_fn, fn)
    if err:
        log.error("Error encountered when listing apps", error=str(err))
        raise Exception("Error encountered when listing apps")
    while resp.has_next():
        async with GlobalRetryController(
            fn_identifier="okta.list_applications"
        ) as retry_controller:
            next_apps, err = await retry_controller(handle_okta_fn, resp.next)
        if err:
            log.error("Error encountered when listing apps", error=str(err))
            raise Exception("Error encountered when listing apps")
        raw_apps.append(next_apps)

    if not raw_apps:
        return []

    tasks = []
    apps = []
    for app_raw in raw_apps:
        app = App(
            id=app_raw.id,
            idp_name=okta_organization.idp_name,
            name=app_raw.label,
            status=app_raw.status,
            extra=dict(
                created=app_raw.created,
            ),
        )
        apps.append(app)
        tasks.append(list_app_user_assignments(okta_organization, app))
        tasks.append(list_app_group_assignments(okta_organization, app))
    app_assignments = await asyncio.gather(*tasks)
    apps_to_return = []
    for app in apps:
        assignments = [a for a in app_assignments if a["app_id"] == app.id]
        for assignment in assignments:
            for user_assignment in assignment.get("user_assignments", []):
                app.assignments.append(Assignment(user=user_assignment))
            for group_assignment in assignment.get("group_assignments", []):
                app.assignments.append(Assignment(group=group_assignment))
        apps_to_return.append(app)
    return apps_to_return


async def update_app_assignments(
    app: App,
    new_assignments: List[Assignment],
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the assignments of a app in Okta.

    Args:
        app (App): The app to update the assignments of.
        new_assignments (List[Assignment]): The new assignments to add to the app.
        okta_organization (OktaOrganization): The Okta organization to update the app in.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    client = await okta_organization.get_okta_client()
    response = []
    # TODO: Only compare user/group assignments, not expires_at
    current_user_assignments = [
        assignment.user for assignment in app.assignments if assignment.user
    ]
    desired_user_assignments = [
        assignment.user for assignment in new_assignments if assignment.user
    ]
    user_assignments_to_unassign = [
        assignment
        for assignment in current_user_assignments
        if assignment not in desired_user_assignments
    ]
    user_assignments_to_assign = [
        assignment
        for assignment in desired_user_assignments
        if assignment not in current_user_assignments
    ]

    current_group_assignments = [
        assignment.group for assignment in app.assignments if assignment.group
    ]
    desired_group_assignments = [
        assignment.group for assignment in new_assignments if assignment.group
    ]
    group_assignments_to_unassign = [
        assignment
        for assignment in current_group_assignments
        if assignment not in desired_group_assignments
    ]

    group_assignments_to_assign = [
        assignment
        for assignment in desired_group_assignments
        if assignment not in current_group_assignments
    ]

    assignments_to_unassign = bool(
        user_assignments_to_unassign or group_assignments_to_unassign
    )
    assignments_to_assign = bool(
        user_assignments_to_assign or group_assignments_to_assign
    )
    if assignments_to_unassign:
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_id=app.id,
                resource_type=app.resource_type,
                attribute="assignments",
                change_summary={
                    "AssignmentsToUnassign": {
                        "user_assignments_to_unassign": user_assignments_to_unassign,
                        "group_assignments_to_unassign": group_assignments_to_unassign,
                    }
                },
                current_value=current_user_assignments + current_group_assignments,
                new_value=desired_user_assignments + desired_group_assignments,
            )
        )

    if assignments_to_assign:
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_id=app.id,
                resource_type=app.resource_type,
                attribute="assignments",
                change_summary={
                    "AssignmentsToAssign": {
                        "user_assignments_to_assign": user_assignments_to_assign,
                        "group_assignments_to_assign": group_assignments_to_assign,
                    }
                },
            )
        )

    if ctx.execute:
        for assignment in user_assignments_to_assign:
            async with GlobalRetryController(
                fn_identifier="okta.get_user"
            ) as retry_controller:
                fn = functools.partial(client.get_user, assignment)
                user_okta, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error("Error retrieving user", user=assignment, **log_params)
                continue
            app_user = models.AppUser({"id": user_okta.id})
            async with GlobalRetryController(
                fn_identifier="okta.assign_user_to_application"
            ) as retry_controller:
                fn = functools.partial(
                    client.assign_user_to_application, app.id, app_user
                )
                _, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error(
                    "Error assigning user to app",
                    user=assignment.user,
                    **log_params,
                )
                continue
        for assignment in group_assignments_to_assign:
            group: Group = await get_group("", assignment, okta_organization)
            if not group:
                log.error("Error retrieving group", group=assignment, **log_params)
                continue
            async with GlobalRetryController(
                fn_identifier="okta.get_group"
            ) as retry_controller:
                fn = functools.partial(client.get_group, group.group_id)
                group_okta, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error("Error retrieving group", group=assignment, **log_params)
                continue
            group_assignment = models.ApplicationGroupAssignment(
                {
                    "id": group_okta.id,
                }
            )
            async with GlobalRetryController(
                fn_identifier="okta.create_application_group_assignment"
            ) as retry_controller:
                fn = functools.partial(
                    client.create_application_group_assignment,
                    app.id,
                    group_okta.id,
                    group_assignment,
                )
                _, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error(
                    "Error assigning group to app",
                    group=assignment.oup,
                    **log_params,
                )
                continue
        for assignment in user_assignments_to_unassign:
            async with GlobalRetryController(
                fn_identifier="okta.get_user"
            ) as retry_controller:
                fn = functools.partial(client.get_user, assignment)
                user_okta, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error("Error retrieving user", user=assignment, **log_params)
                continue
            async with GlobalRetryController(
                fn_identifier="okta.delete_application_user"
            ) as retry_controller:
                fn = functools.partial(
                    client.delete_application_user, app.id, user_okta.id
                )
                _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error(
                    "Error unassigning user from app",
                    user=assignment.user,
                    **log_params,
                )
                continue
        for assignment in group_assignments_to_unassign:
            group: Group = await get_group("", assignment, okta_organization)
            if not group:
                log.error("Error retrieving group", group=assignment, **log_params)
                continue
            async with GlobalRetryController(
                fn_identifier="okta.get_group"
            ) as retry_controller:
                fn = functools.partial(client.get_group, group.group_id)
                group_okta, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error("Error retrieving group", group=assignment, **log_params)
                continue
            async with GlobalRetryController(
                fn_identifier="okta.delete_application_group_assignment"
            ) as retry_controller:
                fn = functools.partial(
                    client.delete_application_group_assignment, app.id, group_okta.id
                )
                _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error(
                    "Error unassigning group from app",
                    group=assignment.group,
                    **log_params,
                )
                continue

    return response


async def update_app_name(
    app: App,
    new_name: str,
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
):
    response: list[ProposedChange] = []
    if app.name == new_name:
        return response
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_type=app.resource_type,
            resource_id=app.resource_id,
            attribute="app_name",
            current_value=app.name,
            new_value=new_name,
        )
    )

    app_model = models.Application(
        {
            "name": new_name,
            "label": new_name,
        }
    )

    if ctx.execute:
        client = await okta_organization.get_okta_client()
        async with GlobalRetryController(
            fn_identifier="okta.update_application"
        ) as retry_controller:
            fn = functools.partial(client.update_application, app.id, app_model)
            _, _, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise ValueError(f"Error updating Okta app: {err}")
    return response


async def maybe_delete_app(
    delete: bool,
    app: App,
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Delete a app in Okta.

    Args:
        app (App): The app to delete.
        okta_organization (OktaOrganization): The Okta organization to delete the group from.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = []
    if not delete:
        return response
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.DELETE,
            resource_id=app.id,
            resource_type=app.resource_type,
            attribute="app",
            change_summary={"app": app.name},
            current_value=app.name,
            new_value=None,
        )
    )
    if ctx.execute:
        client = await okta_organization.get_okta_client()
        async with GlobalRetryController(
            fn_identifier="okta.delete_app"
        ) as retry_controller:
            fn = functools.partial(client.delete_app, app.id)
            _, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise Exception("Error deleting app")
    return response



from __future__ import annotations

import json
import logging
import os

import boto3
import structlog


def pretty_log(logger, method_name, event_dict):
    return_dict = {}

    # Uncomment this if we want to default to single line logging. Can get hard to read with long resources though.
    # if not any(bool(isinstance(value, dict) or isinstance(value, list)) for value in event_dict.values()):
    #     return event_dict

    for key, value in event_dict.items():
        if key != "event":
            key = f"\n  {key}"

        if isinstance(value, dict) or isinstance(value, list):
            value = "\n  ".join(str(json.dumps(value, indent=2)).split("\n"))

        return_dict[key] = value

    return return_dict


LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO")
logging.basicConfig(level=LOG_LEVEL)
structlog.configure(
    processors=[
        pretty_log,
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.processors.TimeStamper("%Y/%m/%d %H:%M:%S", utc=False),
        structlog.dev.ConsoleRenderer(),
    ],
    wrapper_class=structlog.make_filtering_bound_logger(
        logging.getLevelName(LOG_LEVEL)
    ),
    context_class=dict,
    logger_factory=structlog.PrintLoggerFactory(),
    cache_logger_on_first_use=False,
)
log = structlog.get_logger("NoqForm")

if LOG_LEVEL == "DEBUG":
    boto3.set_stream_logger("", "DEBUG")
else:
    default_logging_levels = {
        "boto3": "CRITICAL",
        "boto": "CRITICAL",
        "botocore": "CRITICAL",
    }
    for logger, level in default_logging_levels.items():
        logging.getLogger(logger).setLevel(level)



#!/usr/bin/env python3
from __future__ import annotations

import asyncio
import hashlib
import hmac
import json
import os
import tempfile
import time
from typing import Any, Callable
from urllib.parse import urlparse

import aiohttp
import boto3
import github
import jwt
from botocore.exceptions import ClientError

import iambic.core.utils
import iambic.plugins.v0_1_0.github.github

# from iambic.core.git import get_remote_default_branch
from iambic.core.logger import log
from iambic.plugins.v0_1_0.github.github import (
    HandleIssueCommentReturnCode,
    _handle_detect_changes_from_eventbridge,
    _handle_enforce,
    _handle_expire,
    _handle_import,
    github_app_workflow_wrapper,
    handle_iambic_git_apply,
    handle_iambic_git_plan,
    iambic_app,
)

# FIXME Lambda execution time is at most 15 minutes, and the Github installation token is at most
# 10 min validation period.

# FIXME struct logs is not showing up on lambda cloudwatch logs

# FIXME exception during git-plan is unknown
# /tmp/.iambic/repos/ already exists. This is unexpected.
# This is due to lambda reusing the already running container


def format_github_url(repository_url: str, github_token: str) -> str:
    parse_result = urlparse(repository_url)
    return parse_result._replace(
        netloc="x-access-token:{0}@{1}".format(github_token, parse_result.netloc)
    ).geturl()


def get_app_bearer_token(private_key, app_id) -> str:
    payload = {
        # Issued at time
        "iat": int(time.time()),
        # JWT expiration time (10 minutes maximum)
        "exp": int(time.time()) + 600,
        # GitHub App's identifier
        "iss": app_id,
    }

    # Create JWT
    return jwt.encode(payload, private_key, algorithm="RS256")


def get_app_private_key_as_lambda_context():
    # assuming we are already in an lambda execution context
    secret_name = os.environ["GITHUB_APP_SECRET_KEY_SECRET_ID"]
    region_name = os.environ["AWS_REGION"]

    # Create a Secrets Manager client
    session = boto3.session.Session()
    client = session.client(service_name="secretsmanager", region_name=region_name)

    try:
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    except ClientError as e:
        # For a list of exceptions thrown, see
        # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
        raise e

    # Decrypts secret using the associated KMS key.
    return get_secret_value_response["SecretString"]


def get_app_webhook_secret_as_lambda_context():
    # assuming we are already in an lambda execution context
    secret_name = os.environ["GITHUB_APP_WEBHOOK_SECRET_SECRET_ID"]
    region_name = os.environ["AWS_REGION"]

    # Create a Secrets Manager client
    session = boto3.session.Session()
    client = session.client(service_name="secretsmanager", region_name=region_name)

    try:
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    except ClientError as e:
        # For a list of exceptions thrown, see
        # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
        raise e

    # Decrypts secret using the associated KMS key.
    return get_secret_value_response["SecretString"]


async def _get_installation_token(app_id, installation_id):
    encoded_jwt = get_app_bearer_token(get_app_private_key_as_lambda_context(), app_id)
    access_tokens_url = (
        f"https://api.github.com/app/installations/{installation_id}/access_tokens"
    )
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {encoded_jwt}",
    }
    async with aiohttp.ClientSession() as session:
        async with session.post(access_tokens_url, headers=headers) as resp:
            payload = json.loads(await resp.text())
            installation_token = payload["token"]
            return installation_token


def run_handler(event=None, context=None):
    """
    Default handler for AWS Lambda. It is split out from the actual
    handler so we can also run via IDE run configurations
    """

    # debug
    print(event)

    github_event = event["headers"]["x-github-event"]
    app_id = event["headers"]["x-github-hook-installation-target-id"]
    request_signature = event["headers"]["x-hub-signature-256"].split("=")[
        1
    ]  # the format is in sha256=<sig>

    # verify webhooks security secrets
    payload = event["body"]
    verify_signature(request_signature, payload)

    webhook_payload = json.loads(event["body"])
    installation_id = webhook_payload["installation"]["id"]
    github_override_token = asyncio.run(
        _get_installation_token(app_id, installation_id)
    )

    github_client = github.Github(github_override_token)

    # Handle lambda environment can only write to /tmp and make sure we don't leave previous
    # state on a new function execution
    temp_templates_directory = tempfile.mkdtemp(prefix="lambda")
    os.chdir(
        temp_templates_directory
    )  # the rest of the system seems to use filesystem for stuff
    getattr(iambic_app, "lambda").app.init_plan_output_path()
    getattr(iambic_app, "lambda").app.init_repo_base_path()
    iambic.plugins.v0_1_0.github.github.init_shared_data_directory()
    iambic.core.utils.init_writable_directory()

    f: Callable[[str, github.Github, dict[str, Any]]] = EVENT_DISPATCH_MAP.get(
        github_event
    )
    if f:
        f(github_override_token, github_client, webhook_payload)
    else:
        log.error("no supported handler")
        raise Exception("no supported handler")


def handle_pull_request(
    github_token: str, github_client: github.Github, webhook_payload: dict[str, Any]
) -> None:
    action = webhook_payload["action"]
    if action not in ["opened", "synchronize"]:
        return

    repo_name = webhook_payload["repository"]["full_name"]
    pull_number = webhook_payload["pull_request"]["number"]
    # repository_url_token
    templates_repo = github_client.get_repo(repo_name)
    pull_request = templates_repo.get_pull(pull_number)
    repository_url = webhook_payload["repository"]["clone_url"]
    repo_url = format_github_url(repository_url, github_token)
    pull_request_branch_name = pull_request.head.ref

    if action == "synchronize":
        commits: list[github.Commit.Commit] = list(pull_request.get_commits())
        last_commit_message = commits[-1].commit.message
        if (
            last_commit_message
            == iambic.plugins.v0_1_0.github.github.COMMIT_MESSAGE_FOR_GIT_APPLY_ABSOLUTE_TIME
        ):
            log.info(
                "github_app ignore synchronize event since its likely triggered by itself"
            )
            return
        else:
            log_params = {"last_commit_message": last_commit_message}
            log.info("last known commit message", **log_params)

    return handle_iambic_git_plan(
        None,
        github_client,
        templates_repo,
        pull_request,
        repo_name,
        pull_number,
        pull_request_branch_name,
        repo_url,
        proposed_changes_path=getattr(iambic_app, "lambda").app.PLAN_OUTPUT_PATH,
    )


def handle_issue_comment(
    github_token: str, github_client: github.Github, webhook_payload: dict[str, Any]
) -> HandleIssueCommentReturnCode:
    action = webhook_payload["action"]
    if action != "created":
        return

    comment_body = webhook_payload["comment"]["body"]
    comment_user_login = webhook_payload["comment"]["user"]["login"]
    log_params = {"COMMENT_DISPATCH_MAP_KEYS": COMMENT_DISPATCH_MAP.keys()}
    log.info("COMMENT_DISPATCH_MAP keys", **log_params)
    if comment_body not in COMMENT_DISPATCH_MAP:
        log_params = {"comment_body": comment_body}
        log.error("handle_issue_comment: no op", **log_params)
        return HandleIssueCommentReturnCode.NO_MATCHING_BODY

    # FIXME: Need to find a mechanism to avoid infinite loop
    # the following is a very crude one
    if comment_user_login.endswith("[bot]"):
        # return early
        return HandleIssueCommentReturnCode.UNDEFINED

    repo_name = webhook_payload["repository"]["full_name"]
    pull_number = webhook_payload["issue"]["number"]
    # repository_url_token
    templates_repo = github_client.get_repo(repo_name)
    pull_request = templates_repo.get_pull(pull_number)

    # repo_name is already in the format {repo_owner}/{repo_short_name}
    repository_url = webhook_payload["repository"]["clone_url"]
    repo_url = format_github_url(repository_url, github_token)
    # repository_url_token
    templates_repo = github_client.get_repo(repo_name)
    pull_request = templates_repo.get_pull(pull_number)
    pull_request_branch_name = pull_request.head.ref
    log_params = {"pull_request_branch_name": pull_request_branch_name}
    log.info("PR remote branch name", **log_params)

    comment_func: Callable = COMMENT_DISPATCH_MAP[comment_body]
    return comment_func(
        None,
        github_client,
        templates_repo,
        pull_request,
        repo_name,
        pull_number,
        pull_request_branch_name,
        repo_url,
        proposed_changes_path=getattr(iambic_app, "lambda").app.PLAN_OUTPUT_PATH,
    )


def handle_workflow_run(
    github_token: str, github_client: github.Github, webhook_payload: dict[str, Any]
) -> None:
    action = webhook_payload["action"]
    if action != "requested":
        return

    workflow_path = webhook_payload["workflow_run"]["path"]

    if workflow_path not in WORKFLOW_DISPATCH_MAP:
        log_params = {"workflow_path": workflow_path}
        log.error("handle_workflow_run: no op", **log_params)
        return

    repository_url = webhook_payload["repository"]["clone_url"]
    repo_url = format_github_url(repository_url, github_token)

    repo_name = webhook_payload["repository"]["full_name"]
    templates_repo = github_client.get_repo(repo_name)
    default_branch = templates_repo.default_branch

    workflow_func: Callable = WORKFLOW_DISPATCH_MAP[workflow_path]
    return workflow_func(
        github_client,
        templates_repo,
        repo_name,
        repo_url,
        default_branch,
        proposed_changes_path=getattr(iambic_app, "lambda").app.PLAN_OUTPUT_PATH,
    )


EVENT_DISPATCH_MAP: dict[str, Callable] = {
    "issue_comment": handle_issue_comment,
    "pull_request": handle_pull_request,
    "workflow_run": handle_workflow_run,
}


COMMENT_DISPATCH_MAP: dict[str, Callable] = {
    "iambic git-apply": handle_iambic_git_apply,
    "iambic git-plan": handle_iambic_git_plan,
    "iambic apply": handle_iambic_git_apply,
    "iambic plan": handle_iambic_git_plan,
}

WORKFLOW_DISPATCH_MAP: dict[str, Callable] = {
    ".github/workflows/iambic-enforce.yml": github_app_workflow_wrapper(
        _handle_enforce, "enforce"
    ),
    ".github/workflows/iambic-expire.yml": github_app_workflow_wrapper(
        _handle_expire, "expire"
    ),
    ".github/workflows/iambic-import.yml": github_app_workflow_wrapper(
        _handle_import, "import"
    ),
    ".github/workflows/iambic-detect.yml": github_app_workflow_wrapper(
        _handle_detect_changes_from_eventbridge, "detect"
    ),
}


# Use to verify Github App Webhook Secret Using SHA256
def calculate_signature(webhook_secret: str, payload: str) -> str:
    secret_in_bytes = bytes(webhook_secret, "utf-8")
    digest = hmac.new(
        key=secret_in_bytes, msg=payload.encode("utf-8"), digestmod=hashlib.sha256
    )
    signature = digest.hexdigest()
    return signature


def verify_signature(sig: str, payload: str) -> None:
    good_sig = calculate_signature(get_app_webhook_secret_as_lambda_context(), payload)
    if not hmac.compare_digest(good_sig, sig):
        raise Exception("Bad signature")


if __name__ == "__main__":
    print("not supported")



from __future__ import annotations

import asyncio
from typing import Any, Awaitable, TypeVar

T = TypeVar("T")


async def gather_limit(
    *args: Awaitable[T],
    return_exceptions: bool = False,
    limit: int = -1,
) -> list[Any]:
    """
    (Taken from https://github.com/omnilib/aioitertools/blob/v0.7.1/aioitertools/asyncio.py)
    Like asyncio.gather but with a limit on concurrency.

    Note that all results are buffered.

    If gather is cancelled all tasks that were internally created and still pending
    will be cancelled as well.

    Example::

        futures = [some_coro(i) for i in range(10)]

        results = await gather(*futures, limit=2)
    """

    # For detecting input duplicates and reconciling them at the end
    input_map: dict[Awaitable[T], list[int]] = {}
    # This is keyed on what we'll get back from asyncio.wait
    pos: dict[asyncio.Future[T], int] = {}
    ret: list[Any] = [None] * len(args)

    pending: set[asyncio.Future[T]] = set()
    done: set[asyncio.Future[T]] = set()

    next_arg = 0

    while True:
        while next_arg < len(args) and (limit == -1 or len(pending) < limit):
            # We have to defer the creation of the Task as long as possible
            # because once we do, it starts executing, regardless of what we
            # have in the pending set.
            if args[next_arg] in input_map:
                input_map[args[next_arg]].append(next_arg)
            else:
                # We call ensure_future directly to ensure that we have a Task
                # because the return value of asyncio.wait will be an implicit
                # task otherwise, and we won't be able to know which input it
                # corresponds to.
                task: asyncio.Future[T] = asyncio.ensure_future(args[next_arg])
                pending.add(task)
                pos[task] = next_arg
                input_map[args[next_arg]] = [next_arg]
            next_arg += 1

        # pending might be empty if the last items of args were dupes;
        # asyncio.wait([]) will raise an exception.
        if pending:
            try:
                done, pending = await asyncio.wait(
                    pending, return_when=asyncio.FIRST_COMPLETED
                )
                for x in done:
                    if return_exceptions and x.exception():
                        ret[pos[x]] = x.exception()
                    else:
                        ret[pos[x]] = x.result()
            except asyncio.CancelledError:
                # Since we created these tasks we should cancel them
                for x in pending:
                    x.cancel()
                # we insure that all tasks are cancelled before we raise
                await asyncio.gather(*pending, return_exceptions=True)
                raise

        if not pending and next_arg == len(args):
            break

    for lst in input_map.values():
        for i in range(1, len(lst)):
            ret[lst[i]] = ret[lst[0]]

    return ret



from __future__ import annotations

import pytest
from googleapiclient.errors import UnknownApiNameOrVersion
from mock import MagicMock, patch

from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import (
    GoogleProject,
    GoogleSubject,
    SecretStr,
)


@pytest.fixture
def google_project():
    return GoogleProject(
        project_id="example-ga",
        project_name=None,
        subjects=[
            GoogleSubject(
                domain="example.com",
                service_account="google-groups-test@example.iam.gserviceaccount.com",
            )
        ],
        type="service_account",
        private_key_id="123456",
        private_key=SecretStr("**********"),
        client_email="google-groups-test@example.iam.gserviceaccount.com",
        client_id="123456",
        auth_uri="https://accounts.google.com/o/oauth2/auth",
        token_uri="https://oauth2.googleapis.com/token",
        auth_provider_x509_cert_url="https://www.googleapis.com/oauth2/v1/certs",
        client_x509_cert_url="https://www.googleapis.com/robot/v1/metadata/x509/google-groups-test%40example-ga.iam.gserviceaccount.com",
        variables=[],
        iambic_managed=None,
    )


def test_google_project_str(google_project):
    assert str(google_project) == "example-ga"


@pytest.mark.asyncio
@patch("google.oauth2.service_account.Credentials.from_service_account_info")
async def test_google_project_get_service_connection(
    mock_from_service_account_info, google_project
):
    mock_credentials = MagicMock()
    mock_from_service_account_info.return_value = mock_credentials

    service_conn = await google_project.get_service_connection(
        service_name="admin",
        service_path="directory_v1",
        domain="example.com",
        cache_discovery=False,
    )

    assert service_conn is not None
    mock_from_service_account_info.assert_called_once()


@pytest.mark.asyncio
@patch("google.oauth2.service_account.Credentials.from_service_account_info")
async def test_google_project_get_service_connection_no_subject(
    mock_from_service_account_info, google_project
):
    mock_credentials = MagicMock()
    mock_from_service_account_info.return_value = mock_credentials

    with pytest.raises(
        Exception,
        match=r"Could not find service account for domain non_existent_domain",
    ):
        await google_project.get_service_connection(
            service_name="admin",
            service_path="directory_v1",
            domain="non_existent_domain",
            cache_discovery=False,
        )


@pytest.mark.asyncio
@patch("google.oauth2.service_account.Credentials.from_service_account_info")
async def test_google_project_get_service_connection_unknown_api(
    mock_from_service_account_info, google_project
):
    mock_credentials = MagicMock()
    mock_from_service_account_info.return_value = mock_credentials

    with pytest.raises(UnknownApiNameOrVersion):
        await google_project.get_service_connection(
            service_name="non_existent_service",
            service_path="non_existent_service_v1",
            domain="example.com",
            cache_discovery=False,
        )



from __future__ import annotations

from pydantic import BaseModel, Field

from iambic.core.iambic_plugin import ProviderPlugin
from iambic.plugins.v0_1_0 import PLUGIN_VERSION
from iambic.plugins.v0_1_0.github.handlers import import_github_resources, load


class GithubConfig(BaseModel):
    commit_message_user_name: str = Field(
        default="Iambic Automation", description="Commit message user name"
    )
    commit_message_user_email: str = Field(
        default="iambic-automation@iambic.org", description="Commit message user email"
    )
    commit_message_for_detect: str = Field(
        default="Import changes from detect operation",
        description="Commit message to use during changes through detect operations",
    )
    commit_message_for_import: str = Field(
        default="Import changes from import operation",
        description="Commit message to use during changes through import operations",
    )
    commit_message_for_expire: str = Field(
        default="Periodic Expiration",
        description="Commit message to use during changes through expire operations",
    )
    commit_message_for_git_apply: str = Field(
        default="Replace relative time with absolute time",
        description="Commit message to use during changes through git-apply",
    )


IAMBIC_PLUGIN = ProviderPlugin(
    config_name="github",
    version=PLUGIN_VERSION,
    provider_config=GithubConfig,
    requires_secret=False,
    async_import_callable=import_github_resources,
    async_load_callable=load,
    templates=[],
)



from __future__ import annotations

from collections import defaultdict
from typing import Union

import xxhash

from iambic.core import noq_json as json
from iambic.core.iambic_enum import IambicManaged
from iambic.core.logger import log
from iambic.core.models import AccessModelMixin, BaseModel, BaseTemplate, ProviderChild
from iambic.core.parser import load_templates
from iambic.core.utils import (
    evaluate_on_provider,
    gather_templates,
    get_provider_value,
    is_regex_match,
    sanitize_string,
)


async def get_existing_template_map(
    repo_dir: str, template_type: str, nested: bool = False
) -> dict:
    """Used to keep track of existing templates on import

     Write to the existing file before creating a new one.

    :param repo_dir:
    :param template_type:
    :param nested: If true, will return a map of {template_type: {resource_id: template}}
    :return: {resource_id: template}
    """
    templates = load_templates(await gather_templates(repo_dir, template_type))
    if not nested:
        return {template.resource_id: template for template in templates}

    response = defaultdict(dict)
    for template in templates:
        response[template.template_type][template.resource_id] = template
    return response


def templatize_resource(
    provider_child: ProviderChild, resource, valid_characters_re: str = r"[\w_+=,.@-]"
):
    resource_type = type(resource)

    if isinstance(resource, dict) or isinstance(resource, list):
        resource = json.dumps(resource)
    elif resource_type != str:
        resource = str(resource)

    if variables := getattr(provider_child, "variables", None):
        for var in variables:
            if isinstance(var.value, str):
                var.value = sanitize_string(var.value, valid_characters_re)

            var_key_str = "{{var.{}}}".format(var.key)
            resource = resource.replace(var.value, "{{{}}}".format(var_key_str))

    return (
        json.loads(resource)
        if resource_type == dict or resource_type == list
        else resource_type(resource)
    )


def base_group_int_attribute(
    provider_child_vals: dict,
    provider_child_key_id: str,
) -> dict[int, list[dict[str, str]]]:
    """Groups an int attribute by a shared name across provider children.

    Just call group_int_or_str_attribute

    :param provider_child_vals: dict(resource_val: int = list[str])
    :param provider_child_key_id: The key used to represent the provider child within the resource. For example, in AWS this would be "account_id".
    :return: dict(attribute_val: int = list[dict(provider_child_key_id: str)])
    """
    response: dict[int, list[dict]] = defaultdict(list)

    for provider_child_id, resource_val in provider_child_vals.items():
        response[resource_val].append({provider_child_key_id: provider_child_id})

    return response


async def base_group_str_attribute(
    provider_child_map: dict[str, ProviderChild],
    provider_child_resources: list[dict],
    provider_child_key_id: str,
) -> dict[str, list]:
    """Groups a string attribute by a shared name across provider children

    The ability to pass in and maintain arbitrary keys is necessary for
        parsing resource names related to a provider response

    Call group_int_or_str_attribute instead unless you need to transform this response.
    An example would be grouping role names for generating the template where you need to keep the file_path ref.

    :param provider_child_map: {child_key_id: ProviderChild}
    :param provider_child_resources: list[dict(child_key_id:str, resources=list[dict])]
    :param provider_child_key_id: The key used to represent the provider child within the resource. For example, in AWS this would be "account_id".
    :return: dict(attribute_val: str = list[dict(resource_val: str, provider_child_key_id: str, **)])
    """

    """
    Create map with different representations of a resource value for each account
    (Note that we now only keep the post-templatized version)

    The purpose is to add the 2 ways look-ups are done and maintain o(1) performance.
    The resource_val to the corresponding list element in provider_child_resources[elem]["resources"]
        Under resource_val_map
    The reverse of resource_val_map which is an int representing the elem with a list of all resource_val reprs
        Under elem_resource_val_map
    """
    for provider_child_resource_elem, provider_child_resource in enumerate(
        provider_child_resources
    ):
        provider_child_resources[provider_child_resource_elem][
            "resource_val_map"
        ] = dict()
        provider_child_resources[provider_child_resource_elem][
            "elem_resource_val_map"
        ] = dict()
        provider_child = provider_child_map[
            provider_child_resource[provider_child_key_id]
        ]
        for resource_elem, resource in enumerate(provider_child_resource["resources"]):
            provider_child_resource["resources"][resource_elem][
                provider_child_key_id
            ] = provider_child_resource[provider_child_key_id]
            resource_val = resource["resource_val"]
            templatized_resource_val = templatize_resource(provider_child, resource_val)

            # note about the decision we only kept the post-templatized version
            # we aggressively templatized the incoming information.
            # a note for future reader: if you attempt to conditional decide
            # to templatize or not, you have to be careful regarding greedy
            # algorithm that does not arrive at the same termination state.
            # Concretely, if a literal can both be repeated across accounts
            # or templatized across accounts, greedy algorithm may reach
            # different states.

            provider_child_resources[provider_child_resource_elem]["resource_val_map"][
                templatized_resource_val
            ] = resource_elem
            provider_child_resources[provider_child_resource_elem][
                "elem_resource_val_map"
            ][resource_elem] = [templatized_resource_val]

    grouped_resource_map = defaultdict(
        list
    )  # val:str = list(dict(name: str, path: str, account_id: str))
    # Iterate everything looking for shared names across provider children
    for outer_elem in range(len(provider_child_resources)):
        for resource_val, outer_resource_elem in provider_child_resources[outer_elem][
            "resource_val_map"
        ].items():
            if outer_resource_elem is None:  # It hit on something already
                continue
            for inner_elem in range(outer_elem + 1, len(provider_child_resources)):
                if (
                    inner_resource_elem := provider_child_resources[inner_elem][
                        "resource_val_map"
                    ].get(resource_val)
                ) is not None:
                    if not grouped_resource_map.get(resource_val):
                        # Null out the outer_resource_elem in all the places
                        for rn in provider_child_resources[outer_elem][
                            "elem_resource_val_map"
                        ][outer_resource_elem]:
                            provider_child_resources[outer_elem]["resource_val_map"][
                                rn
                            ] = None
                        provider_child_resources[outer_elem]["elem_resource_val_map"][
                            outer_resource_elem
                        ] = []

                        grouped_resource_map[resource_val] = [
                            provider_child_resources[inner_elem]["resources"][
                                inner_resource_elem
                            ],
                            provider_child_resources[outer_elem]["resources"][
                                outer_resource_elem
                            ],
                        ]
                    else:
                        grouped_resource_map[resource_val].append(
                            provider_child_resources[inner_elem]["resources"][
                                inner_resource_elem
                            ]
                        )

                    for rn in provider_child_resources[inner_elem][
                        "elem_resource_val_map"
                    ][inner_resource_elem]:
                        provider_child_resources[inner_elem]["resource_val_map"][
                            rn
                        ] = None
                    provider_child_resources[inner_elem]["elem_resource_val_map"][
                        inner_resource_elem
                    ] = []

    # Set the remaining attributes unique attributes
    for provider_child_resource in provider_child_resources:
        for elem, resource_vals in provider_child_resource[
            "elem_resource_val_map"
        ].items():
            if not resource_vals:
                continue
            elif len(resource_vals) == 1:
                resource_val = resource_vals[0]
            else:
                # Take priority over raw output
                resource_val = [rv for rv in resource_vals if "{{" not in rv][0]

            provider_child_resource["resources"][elem][
                provider_child_key_id
            ] = provider_child_resource[provider_child_key_id]
            grouped_resource_map[resource_val] = [
                provider_child_resource["resources"][elem]
            ]

    return grouped_resource_map


async def base_group_dict_attribute(
    provider_child_map: dict[str, ProviderChild],
    provider_child_resources: list[dict],
    provider_child_key_id: str,
    included_children_key: str,
    prefer_templatized=False,  # clarification that this keyword parameter has no impact at the moment
) -> list[dict]:
    """Groups an attribute that is a dict or list of dicts with matching provider children.

    Call group_dict_attribute instead unless you need to transform this response.
    An example would be tags which also contain access_rules.

    :param provider_child_map: {child_key_id: ProviderChild}
    :param provider_child_resources: list[dict(child_key_id:str, resources=list[dict])]
    :param provider_child_key_id: The key used to represent the provider child within the resource. For example, in AWS this would be "account_id".
    :param included_children_key: The key on the template used to represent the provider children the resource applies to. For example, in AWS this would be "included_accounts"..    :return: list[dict(included_accounts: str, resource_val=list[dict]|dict)]
    """
    """
    Create map with different representations of a resource value for each provider child

    Create a resource_hash to the corresponding list element in provider_child_resources[elem]["resources"]
        Under resource_hash_map
    Create a reverse of resource_hash_map which is an int representing the elem with a list of all resource_hash reprs
        Under elem_resource_hash_map
    """
    hash_map = dict()

    for provider_child_resource_elem, provider_child_resource in enumerate(
        provider_child_resources
    ):
        provider_child_resources[provider_child_resource_elem][
            "resource_hash_map"
        ] = dict()
        provider_child_resources[provider_child_resource_elem][
            "elem_resource_hash_map"
        ] = dict()
        aws_account = provider_child_map[provider_child_resource[provider_child_key_id]]
        for resource_elem, resource in enumerate(provider_child_resource["resources"]):
            provider_child_resource["resources"][resource_elem][
                provider_child_key_id
            ] = provider_child_resource[provider_child_key_id]
            # Set raw dict
            resource_hash = xxhash.xxh32(
                json.dumps(resource["resource_val"])
            ).hexdigest()
            hash_map[resource_hash] = resource["resource_val"]
            # Set dict with attempted interpolation
            templatized_dict = templatize_resource(
                aws_account, resource["resource_val"]
            )
            templatized_resource_hash = xxhash.xxh32(
                json.dumps(templatized_dict)
            ).hexdigest()
            hash_map[templatized_resource_hash] = templatized_dict
            # Define resource hash mappings
            provider_child_resources[provider_child_resource_elem]["resource_hash_map"][
                resource_hash
            ] = resource_elem
            provider_child_resources[provider_child_resource_elem][
                "elem_resource_hash_map"
            ][resource_elem] = [resource_hash]
            if templatized_resource_hash != resource_hash:
                provider_child_resources[provider_child_resource_elem][
                    "resource_hash_map"
                ][templatized_resource_hash] = resource_elem
                provider_child_resources[provider_child_resource_elem][
                    "elem_resource_hash_map"
                ][resource_elem].append(templatized_resource_hash)

    grouped_resource_map = (
        dict()
    )  # val:str = list(dict(name: str, path: str, provider_child_key_id: str))
    # Iterate everything looking for shared names across aws_accounts
    for outer_elem in range(len(provider_child_resources)):
        for resource_hash, outer_resource_elem in provider_child_resources[outer_elem][
            "resource_hash_map"
        ].items():
            if outer_resource_elem is None:  # It hit on something already
                continue
            for inner_elem in range(outer_elem + 1, len(provider_child_resources)):
                if (
                    inner_resource_elem := provider_child_resources[inner_elem][
                        "resource_hash_map"
                    ].get(resource_hash)
                ) is not None:
                    if not grouped_resource_map.get(resource_hash):
                        grouped_resource_map[resource_hash] = {
                            "resource_val": hash_map[resource_hash],
                            included_children_key: [
                                provider_child_resources[inner_elem][
                                    provider_child_key_id
                                ],
                                provider_child_resources[outer_elem][
                                    provider_child_key_id
                                ],
                            ],
                        }

                        # Null out the outer_resource_elem in all the places
                        for rn in provider_child_resources[outer_elem][
                            "elem_resource_hash_map"
                        ][outer_resource_elem]:
                            provider_child_resources[outer_elem]["resource_hash_map"][
                                rn
                            ] = None
                        provider_child_resources[outer_elem]["elem_resource_hash_map"][
                            outer_resource_elem
                        ] = []
                    else:
                        grouped_resource_map[resource_hash][
                            included_children_key
                        ].append(
                            provider_child_resources[inner_elem][provider_child_key_id]
                        )

                    # Null out the inner_resource_elem in all the places
                    for rn in provider_child_resources[inner_elem][
                        "elem_resource_hash_map"
                    ][inner_resource_elem]:
                        provider_child_resources[inner_elem]["resource_hash_map"][
                            rn
                        ] = None
                    provider_child_resources[inner_elem]["elem_resource_hash_map"][
                        inner_resource_elem
                    ] = []

    # Set the remaining attributes unique attributes
    for provider_child_resource in provider_child_resources:
        for elem, resource_hashes in provider_child_resource[
            "elem_resource_hash_map"
        ].items():
            if not resource_hashes:
                continue
            elif len(resource_hashes) == 1:
                resource_hash = resource_hashes[0]
            else:
                # note about the decision we prefer post-templatized version
                # we aggressively templatized the incoming information.
                # a note for future reader: if you attempt to conditional decide
                # to templatize or not, you have to be careful regarding greedy
                # algorithm that does not arrive at the same termination state.
                # Concretely, if a literal can both be repeated across accounts
                # or templatized across accounts, greedy algorithm may reach
                # different states.
                resource_hash = resource_hashes[
                    -1
                ]  # take the last one because it's the templatized version

            grouped_resource_map[resource_hash] = {
                "resource_val": hash_map[resource_hash],
                included_children_key: [provider_child_resource[provider_child_key_id]],
            }

    return list(grouped_resource_map.values())


async def set_included_provider_children_for_grouped_attribute(
    provider_child_map: dict[str, ProviderChild],
    number_of_children_resource_on: int,
    provider_child_key_id: str,
    included_children_key: str,
    grouped_attribute,
) -> Union[list, dict]:
    """Takes a grouped attribute and formats its included aws_accounts to * or a list of account names

    :param provider_child_map: {child_key_id: ProviderChild}
    :param number_of_children_resource_on:
    :param provider_child_key_id: The key used to represent the provider child within the resource. For example, in AWS this would be "account_id".
    :param included_children_key: The key on the template used to represent the provider children the resource applies to. For example, in AWS this would be "included_accounts"..
    :param grouped_attribute:
    :return:
    """
    if isinstance(grouped_attribute, dict):  # via base_group_str_attribute
        for k, resource_vals in grouped_attribute.items():
            if len(resource_vals) == number_of_children_resource_on:
                included_children = ["*"]
            else:
                included_children = [
                    provider_child_map[rv[provider_child_key_id]].preferred_identifier
                    for rv in resource_vals
                ]
            grouped_attribute[k] = included_children

        return grouped_attribute

    elif isinstance(grouped_attribute, list):  # Generated via base_group_dict_attribute
        for elem in range(len(grouped_attribute)):
            if (
                len(grouped_attribute[elem][included_children_key])
                == number_of_children_resource_on
            ):
                grouped_attribute[elem][included_children_key] = ["*"]
            else:
                included_children = [
                    provider_child_map[rv].preferred_identifier
                    for rv in grouped_attribute[elem][included_children_key]
                ]
                grouped_attribute[elem][included_children_key] = included_children

        return grouped_attribute


async def group_int_or_str_attribute(
    provider_child_map: dict[str, ProviderChild],
    number_of_children_resource_on: int,
    provider_child_resources: Union[dict, list[dict]],
    provider_child_key_id: str,
    included_children_key: str,
    key: Union[int, str],
) -> Union[int, str, list[dict]]:
    """Groups an attribute by provider child, formats the attribute and normalizes the included provider children.

    :param provider_child_map: {child_key_id: ProviderChild}
    :param number_of_children_resource_on:
    :param provider_child_resources: list[dict(child_key_id:str, resources=list[dict])]
    :param provider_child_key_id: The key used to represent the provider child within the resource. For example, in AWS this would be "account_id".
    :param included_children_key: The key on the template used to represent the provider children the resource applies to. For example, in AWS this would be "included_accounts"..
    :param key: Used to form the list[dict] response when there are multiple values for the attribute.
    :return:
    """
    if isinstance(provider_child_resources, list):
        grouped_attribute = await base_group_str_attribute(
            provider_child_map, provider_child_resources, provider_child_key_id
        )
    else:
        grouped_attribute = base_group_int_attribute(
            provider_child_resources, provider_child_key_id
        )

    if len(grouped_attribute) == 1:
        return list(grouped_attribute.keys())[0]

    response = []
    grouped_attribute = await set_included_provider_children_for_grouped_attribute(
        provider_child_map,
        number_of_children_resource_on,
        provider_child_key_id,
        included_children_key,
        grouped_attribute,
    )

    for resource_val, included_accounts in grouped_attribute.items():
        if included_accounts[0] == "*":
            response.append({key: resource_val})
        else:
            response.append({key: resource_val, "included_accounts": included_accounts})

    return response


async def group_dict_attribute(
    provider_child_map: dict[str, ProviderChild],
    number_of_children_resource_on: int,
    provider_child_resources: list[dict],
    provider_child_key_id: str,
    included_children_key: str,
    is_dict_attr: bool = True,
    prefer_templatized: bool = False,
) -> Union[dict, list[dict]]:
    """Groups an attribute by aws_accounts, formats the attribute and normalizes the included aws_accounts.

    :param provider_child_map: {child_key_id: ProviderChild}
    :param number_of_children_resource_on:
    :param provider_child_resources: list[dict(child_key_id:str, resources=list[dict])]
    :param provider_child_key_id: The key used to represent the provider child within the resource. For example, in AWS this would be "account_id".
    :param included_children_key: The key on the template used to represent the provider children the resource applies to. For example, in AWS this would be "included_accounts"..
    :param is_dict_attr: If false and only one hit, still return as a list. Useful for things like inline_policies with a list of dicts.
    :return:
    """
    response = []
    grouped_attributes = await set_included_provider_children_for_grouped_attribute(
        provider_child_map,
        number_of_children_resource_on,
        provider_child_key_id,
        included_children_key,
        (
            await base_group_dict_attribute(
                provider_child_map,
                provider_child_resources,
                provider_child_key_id,
                included_children_key,
                prefer_templatized=prefer_templatized,
            )
        ),
    )

    if len(grouped_attributes) == 1 and is_dict_attr:
        attr_val = grouped_attributes[0]["resource_val"]
        included_children = grouped_attributes[0][included_children_key]
        if included_children != ["*"]:
            attr_val[included_children_key] = included_children

        return attr_val

    for grouped_attr in grouped_attributes:
        attr_val = grouped_attr["resource_val"]
        included_children = grouped_attr[included_children_key]
        if included_children != ["*"]:
            attr_val[included_children_key] = included_children

        response.append(attr_val)

    return response


def create_or_update_template(
    file_path,
    existing_template_map,
    identifier,
    template_cls,
    template_params,
    properties,
    all_provider_children: list[ProviderChild],
):
    new_template = template_cls(
        file_path=file_path,
        properties=properties,
        **template_params,
    )

    # iambic-specific knowledge requires us to load the existing template
    # because it will not be reflected by AWS API.
    if existing_template := existing_template_map.get(identifier, None):
        if existing_template.iambic_managed == IambicManaged.ENFORCED:
            # If the template is marked as ENFORCED, we should not update it during import.
            return
        merged_template = merge_model(
            new_template, existing_template, all_provider_children
        )

        try:
            merged_template.write()
            return merged_template
        except Exception as err:
            log.exception(
                f"Unable to update {template_cls} template.",
                error=str(err),
                template_params=template_params,
            )
            raise
    else:
        try:
            new_template.write()
            return new_template
        except Exception as err:
            log.exception(
                f"Unable to create {template_cls} template.",
                error=str(err),
                template_params=template_params,
            )
            raise


def get_resource_id_to_model_map(models: list[BaseModel]) -> dict[str, BaseModel]:
    resource_id_to_model_map = {}
    for existing_model in models:
        existing_resource_id = None
        try:
            existing_resource_id = existing_model.resource_id
        except NotImplementedError:
            pass
        if existing_resource_id:
            resource_id_to_model_map[existing_resource_id] = existing_model
    return resource_id_to_model_map


def sync_access_model_scope(
    source_access_model: AccessModelMixin, destination_access_model: AccessModelMixin
) -> tuple[AccessModelMixin, AccessModelMixin]:
    destination_access_model.set_included_children(
        list(sorted(set(source_access_model.included_children)))
    )
    destination_access_model.set_excluded_children(
        list(sorted(set(source_access_model.excluded_children)))
    )
    destination_access_model.set_included_parents(
        list(sorted(set(source_access_model.included_parents)))
    )
    destination_access_model.set_excluded_parents(
        list(sorted(set(source_access_model.excluded_parents)))
    )
    return source_access_model, destination_access_model


def update_access_attributes(
    new_model: AccessModelMixin,
    existing_model: AccessModelMixin,
    all_provider_children: list[ProviderChild],
) -> tuple[AccessModelMixin, AccessModelMixin]:
    """
        Syncs the access attributes of the new model with the existing model.

        How it works:
            For each included account in the new value:
            Check if the account would be excluded by the existing value
                Either implicitly or explicitly using rule weighting
            If it would be excluded, add it to included_accounts
                and remove it from excluded_accounts if the full provider name is defined in the excluded_accounts.
                For example, if the excluded_accounts is "prod*" and the account name is prod-1234, the rule is unchanged.
                However, if the excluded_accounts is "prod-1234" the rule will be removed.

    for every known account that is not in included_accounts:
        Check if the account would be included by the existing value
        If it would be included, add it to excluded_accounts

        :param new_model: The new model to sync with the existing model.
        :param existing_model: The existing model to sync with the new model.
        :param all_provider_children: All the children of the provider.
        :return: The new model and the existing model.
    """
    if "*" in new_model.included_children:
        new_model, existing_model = sync_access_model_scope(new_model, existing_model)
        return new_model, existing_model
    else:
        for child in all_provider_children:
            currently_evaluated = evaluate_on_provider(existing_model, child, False)
            evaluated_on_new_model = bool(
                child.preferred_identifier in new_model.included_children
            )
            if evaluated_on_new_model and not currently_evaluated:
                if (
                    child.parent_id
                    and child.parent_id in existing_model.excluded_parents
                ):
                    # Unable to preserve original rules because we may include an unwanted child
                    existing_model, new_model = sync_access_model_scope(
                        existing_model, new_model
                    )
                    return new_model, existing_model

                excluded_children = [
                    ec
                    for ec in existing_model.excluded_children
                    if ec not in child.all_identifiers
                ]
                if excluded_children != existing_model.excluded_children:
                    existing_model.set_excluded_children(excluded_children)
                else:
                    existing_model.included_children.append(child.preferred_identifier)

            if not evaluated_on_new_model and currently_evaluated:
                # If the child was explicitly defined in included_children then remove it
                #   because it is no longer included.
                # If it is defined as part of a wildcard
                #   then we need to preserve the rule but add it to excluded children.
                included_children = [
                    ic
                    for ic in existing_model.included_children
                    if ic not in child.all_identifiers
                ]
                if included_children != existing_model.included_children:
                    existing_model.set_included_children(included_children)
                else:
                    existing_model.excluded_children.append(child.preferred_identifier)

    existing_model, new_model = sync_access_model_scope(existing_model, new_model)
    return new_model, existing_model


def sort_access_models_by_included_children(
    access_models: list[AccessModelMixin], as_most_specific: bool = True
) -> list[AccessModelMixin]:
    return sorted(
        access_models,
        key=lambda x: max(len(child) for child in x.included_children),
        reverse=as_most_specific,
    )


def resolve_model_orphaned_children(
    new_model: AccessModelMixin,
    merged_model_list: list[AccessModelMixin],
    resolved_children: set[str],
    provider_child_map: dict[str, ProviderChild],
) -> tuple[list[AccessModelMixin], set]:
    """
    Attempts to attach "orphaned" children to a model with matching IAMbic metadata.
    This is because the metadata is the only preserved attribute when merging models.

    These children are ones that didn't hit on an existing model's included_children
    """
    all_provider_children = list(provider_child_map.values())

    for included_child in new_model.included_children:
        provider_child = provider_child_map.get(included_child)
        if not provider_child:
            continue

        if any(is_regex_match(included_child, child) for child in resolved_children):
            continue

        for elem, matching_model in enumerate(merged_model_list):
            if any(
                is_regex_match(child, provider_child.preferred_identifier)
                for child in matching_model.excluded_children
            ):
                # Don't merge the child if the model excludes it explicitly
                continue

            """
            If the model does not have iambic_specific_knowledge
                or the model has iambic_specific_knowledge
                    and the new model has the same values for the attributes
                attempt to merge, if successful update the merged_model
            """
            if not getattr(matching_model, "iambic_specific_knowledge", None) or all(
                getattr(new_model, attr, None) == getattr(matching_model, attr, None)
                for attr in matching_model.iambic_specific_knowledge()
            ):
                resolved_children.add(included_child)
                merged_model = merge_model(
                    new_model, matching_model, all_provider_children
                )
                if merged_model:
                    merged_model_list[elem] = merged_model
                    break

    return merged_model_list, resolved_children


def merge_access_model_list(
    new_list: list[AccessModelMixin],
    existing_list: list[AccessModelMixin],
    all_provider_children: list[ProviderChild],
) -> list[AccessModelMixin]:
    """
    If the field is a list of objects that inherit from AccessModel:
    Attempt to resolve the matching model between the 2 lists
        If found, refer to "if the field inherits from AccessModel"
        If not found in the existing value, add the new model to the list
        If not found in the new value, remove the existing model from the list
    """
    merged_list = []
    provider_child_map = {
        child.preferred_identifier: child for child in all_provider_children
    }
    for new_model_index, new_model in enumerate(new_list):
        matching_existing_models = [
            existing_model
            for existing_model in existing_list
            if (existing_model.resource_id == new_model.resource_id)
            and new_model.resource_id  # we cannot match on empty string like value
        ]
        if not matching_existing_models:
            # #attempt to carry over previous list's iambic specific information for the merged_value
            if new_model_index < len(existing_list):
                existing_model = existing_list[new_model_index]
                merged_list.append(
                    # it's important NOT to have merge_model handle update_access_attribute
                    # because it knows synching between existing and incoming. In this portion
                    # of merge_access_model_list, merge_access_model_list is driving everything
                    # about access model
                    merge_model(
                        new_model,
                        existing_model,
                        all_provider_children,
                        should_update_access_attributes=False,
                    ),
                )
            else:
                merged_list.append(new_model)
        elif new_model.included_children == ["*"]:
            # Find the least specific
            matching_existing_models = sort_access_models_by_included_children(
                matching_existing_models, False
            )
            merged_model = merge_model(
                new_model, matching_existing_models[0], all_provider_children
            )
            if merged_model:
                merged_list.append(merged_model)
        else:
            # Find the most specific
            matching_existing_models = sort_access_models_by_included_children(
                matching_existing_models
            )
            """
            Attempt to find in sub_merged_model_list
            if found, skip, add to resolved_children
            if not found, attempt to find in matching_existing_models
            if found, merge, add to sub_merged_model_list, add to resolved_children
            if not found, continue
            Remove all resolved_children from new_model.included_children
            if new model included_children is not empty, add to merged_list
            """
            sub_merged_model_list = []
            resolved_children = set()

            for included_child in new_model.included_children:
                provider_child = provider_child_map.get(included_child)
                if not provider_child:
                    continue

                if sub_merged_model_list:
                    if get_provider_value(
                        sub_merged_model_list, provider_child.all_identifiers
                    ):
                        resolved_children.add(included_child)
                        continue

                existing_model = get_provider_value(
                    matching_existing_models, provider_child.all_identifiers
                )
                if existing_model:
                    resolved_children.add(included_child)
                    merged_model = merge_model(
                        new_model, existing_model, all_provider_children
                    )
                    if merged_model:
                        sub_merged_model_list.append(merged_model)

            # Cannot short circuit simply by resolved_children and
            # included_children length because included_children
            # may not have been expanded (regex like prod*)
            if sub_merged_model_list:
                (
                    sub_merged_model_list,
                    resolved_children,
                ) = resolve_model_orphaned_children(
                    new_model,
                    sub_merged_model_list,
                    resolved_children,
                    provider_child_map,
                )

            merged_list.extend(sub_merged_model_list)
            new_model.set_included_children(
                [
                    child
                    for child in new_model.included_children
                    if "*" not in child and child not in resolved_children
                ]
            )
            if new_model.included_children:
                # Add new_model containing children that were not resolved and couldn't be attached to a model
                merged_list.append(new_model)

    return merged_list


def merge_model_list(
    new_list: list[BaseModel],
    existing_list: list[BaseModel],
    all_provider_children: list[ProviderChild],
) -> list:
    existing_resource_id_to_model_map = get_resource_id_to_model_map(existing_list)
    merged_list = []
    for new_index, new_model in enumerate(new_list):
        new_model_resource_id = None
        try:
            new_model_resource_id = new_model.resource_id
        except NotImplementedError:
            pass
        if (
            new_model_resource_id
            and new_model_resource_id in existing_resource_id_to_model_map
        ):
            existing_model = existing_resource_id_to_model_map[new_model_resource_id]
            merged_list.append(
                merge_model(new_model, existing_model, all_provider_children)
            )
        elif not new_model_resource_id and new_index < len(existing_list):
            # when we cannot use
            existing_model = existing_list[new_index]
            merged_list.append(
                merge_model(new_model, existing_model, all_provider_children)
            )
        else:
            merged_list.append(new_model.copy())
    return merged_list


def merge_model(
    new_model: BaseModel,
    existing_model: BaseModel,
    all_provider_children: list[ProviderChild],
    should_update_access_attributes=True,
) -> Union[BaseModel, list[BaseModel], None]:
    """
    Update the metadata of the new IAMbic model using the existing model.
    The merging process supports merging lists of objects that inherit from the AccessModelMixin, as well as
    merging nested models that inherit from IAMbic's BaseModel.
    This is to preserve metadata on import.

    Args:
    - new_model (BaseModel): The incoming IAMbic BaseModel to be merged with the existing model
    - existing_model (BaseModel): The IAMbic BaseModel to be used to set the new_model's metadata
    - all_provider_children (list[ProviderChild]): A list of provider children used to resolve conflicts when merging
        lists of objects that inherit from the AccessModelMixin

    Returns:
    - Union[BaseModel, list[BaseModel], None]: The merged IAMbic model,
        or a list of merged IAMbic models,
        or None if the new model was set to None

    """
    if new_model is None:
        # The attribute was set to None
        if existing_model:
            log.warn(
                "merge_model: the incoming value is None when existing value is not None"
            )
        return new_model

    merged_model = existing_model.copy()
    iambic_fields = existing_model.metadata_iambic_fields
    field_names = new_model.__fields__.keys()

    if (
        isinstance(merged_model, AccessModelMixin)
        and isinstance(new_model, AccessModelMixin)
        and should_update_access_attributes
    ):
        """
        If the field is a list of objects that inherit from AccessModel:
        Attempt to resolve the matching model between the 2 lists
            If found, refer to "if the field inherits from AccessModel"
            If not found in the existing value, add the new model to the list
            If not found in the new value, remove the existing model from the list
        """
        new_model, merged_model = update_access_attributes(
            new_model, merged_model, all_provider_children
        )

    if isinstance(new_model, list) and not isinstance(merged_model, list):
        """
        If the new value is a list while the existing value is not:
            Cast existing model to a list and have merge_access_model_list attempt to resolve the matching model
        """
        existing_model = [existing_model]
        merged_model = [merged_model]

    elif isinstance(merged_model, list) and not isinstance(new_model, list):
        """
        If the existing value is a list while the new value is not:
            Cast new model to a list and have merge_access_model_list attempt to resolve the matching model
        """
        new_model = [new_model]

    for key in field_names:
        new_value = getattr(new_model, key)
        value_as_list = isinstance(new_value, list)
        existing_value = getattr(existing_model, key)
        if isinstance(existing_value, list):
            if len(existing_value) > 0:
                inner_element = existing_value[0]

                if isinstance(inner_element, AccessModelMixin):
                    """
                    If the field is a list of objects that inherit from the AccessModelMixin:
                    Attempt to resolve the matching model between the 2 lists
                        If found, refer to "if the field inherits from AccessModel"
                        If not found in the existing value, add the new model to the list
                        If not found in the new value, remove the existing model from the list
                    """

                    # type conversion due to field allow mixed types
                    if isinstance(new_value, str):
                        _cls = type(inner_element)
                        new_value = [_cls.new_instance_from_string(new_value)]
                        value_as_list = True  # because cast it into a list
                    elif not isinstance(new_value, list):
                        new_value = [new_value]

                    new_value = merge_access_model_list(
                        new_value, existing_value, all_provider_children
                    )
                    setattr(
                        merged_model, key, new_value if value_as_list else new_value[0]
                    )
                elif isinstance(inner_element, BaseModel):
                    new_value = merge_model_list(
                        new_value, existing_value, all_provider_children
                    )
                    setattr(
                        merged_model, key, new_value if value_as_list else new_value[0]
                    )
                else:
                    setattr(merged_model, key, new_value)
            else:
                setattr(merged_model, key, new_value)
        elif isinstance(existing_value, BaseModel):
            if isinstance(new_value, BaseModel):
                setattr(
                    merged_model,
                    key,
                    merge_model(new_value, existing_value, all_provider_children),
                )
            elif isinstance(new_value, list):
                # handling conversion from existing BaseModel type to incoming list type
                setattr(
                    merged_model,
                    key,
                    merge_model_list(
                        new_value, [existing_value], all_provider_children
                    ),
                )
            elif new_value is None:
                # cloud is represented by None, local is a BaseModel.
                # this will only work if the local BaseModel does not carry
                # any local metadata that needs to survive outside of cloud.
                setattr(merged_model, key, new_value)
            else:
                raise TypeError(
                    f"Type of {type(new_value)} is not supported. Please file a github issue"
                )
        elif key not in iambic_fields:
            setattr(merged_model, key, new_value)
    return merged_model


def delete_orphaned_templates(
    existing_templates: list[BaseTemplate], resource_ids: set[str]
):
    """
    Delete templates that were not found in the latest import for a single template type

    Args:
    - existing_templates (list[BaseTemplate]): List of templates that were already in IAMbic
    - resource_ids (set[str]): The set of resource ids that were found in the latest import
    """
    for existing_template in existing_templates:
        if existing_template.resource_id not in resource_ids:
            if existing_template.iambic_managed == IambicManaged.ENFORCED:
                # If the template is marked as ENFORCED, we should not delete it.
                continue
            log.warning(
                "Removing template that references deleted resource",
                resource_type=existing_template.resource_type,
                resource_id=existing_template.resource_id,
            )
            existing_template.delete()



from __future__ import annotations

import asyncio
from itertools import chain

from botocore.exceptions import ClientError
from deepdiff import DeepDiff

from iambic.core import noq_json as json
from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import NoqSemaphore, aio_wrapper, plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call, paginated_search


async def list_managed_policy_versions(iam_client, policy_arn: str) -> list[dict]:
    return (
        await boto_crud_call(iam_client.list_policy_versions, PolicyArn=policy_arn)
    ).get("Versions", [])


async def list_managed_policy_tags(iam_client, policy_arn: str) -> list[dict]:
    return await paginated_search(
        iam_client.list_policy_tags, "Tags", PolicyArn=policy_arn
    )


async def get_managed_policy_version_doc(
    iam_client, policy_arn: str, version_id: str, **kwargs
) -> dict:
    return (
        (
            await boto_crud_call(
                iam_client.get_policy_version,
                PolicyArn=policy_arn,
                VersionId=version_id,
            )
        )
        .get("PolicyVersion", {})
        .get("Document", {})
    )


async def get_managed_policy_attachments(iam_client, policy_arn: str):
    """
    Get a list of all entities that have the managed policy attached.

    return:
        {
            'PolicyGroups': [{'GroupName': 'string', 'GroupId': 'string'}],
            'PolicyUsers': [{'UserName': 'string', 'UserId': 'string'}],
            'PolicyRoles': [{'RoleName': 'string', 'RoleId': 'string'}]
        }
    """
    return await paginated_search(
        iam_client.list_entities_for_policy,
        response_keys=["PolicyGroups", "PolicyRoles", "PolicyUsers"],
        retain_key=True,
        PolicyArn=policy_arn,
    )


async def get_managed_policy(iam_client, policy_arn: str, **kwargs) -> dict:
    try:
        response = (
            await boto_crud_call(iam_client.get_policy, PolicyArn=policy_arn)
        ).get("Policy", {})
        if response:
            response["PolicyDocument"] = await get_managed_policy_version_doc(
                iam_client, policy_arn, response.pop("DefaultVersionId")
            )
    except ClientError as err:
        if err.response["Error"]["Code"] == "NoSuchEntity":
            response = {}
        else:
            raise

    return response


def get_oldest_policy_version_id(policy_versions: list[dict]) -> str:
    policy_versions = sorted(policy_versions, key=lambda version: version["CreateDate"])
    if not policy_versions[0]["IsDefaultVersion"]:
        return policy_versions[0]["VersionId"]
    elif len(policy_versions) > 1:
        return policy_versions[1]["VersionId"]


async def list_managed_policies(
    iam_client,
    scope: str = "Local",
    only_attached: bool = False,
    path_prefix: str = "/",
    policy_usage_filter: str = None,
):
    get_managed_policy_semaphore = NoqSemaphore(get_managed_policy, 50)
    list_policy_kwargs = dict(
        Scope=scope,
        OnlyAttached=only_attached,
        PathPrefix=path_prefix,
    )
    if policy_usage_filter:
        list_policy_kwargs["PolicyUsageFilter"] = policy_usage_filter

    managed_policies = await paginated_search(
        iam_client.list_policies, response_key="Policies", **list_policy_kwargs
    )
    return await get_managed_policy_semaphore.process(
        [
            {"iam_client": iam_client, "policy_arn": policy["Arn"]}
            for policy in managed_policies
        ]
    )


async def delete_managed_policy(iam_client, policy_arn: str, log_params: dict):
    policy_attachments = await get_managed_policy_attachments(iam_client, policy_arn)
    policy_versions = await list_managed_policy_versions(iam_client, policy_arn)
    tasks = []

    for detachment_type in ["User", "Role", "Group"]:
        for entity in policy_attachments[f"Policy{detachment_type}s"]:
            tasks.append(
                boto_crud_call(
                    getattr(iam_client, f"detach_{detachment_type.lower()}_policy"),
                    PolicyArn=policy_arn,
                    **{f"{detachment_type}Name": entity[f"{detachment_type}Name"]},
                )
            )

    if len(policy_versions) > 1:
        for version in policy_versions:
            if version["IsDefaultVersion"]:
                continue

            tasks.append(
                boto_crud_call(
                    iam_client.delete_policy_version,
                    PolicyArn=policy_arn,
                    VersionId=version["VersionId"],
                )
            )

    log.debug(
        "Detaching managed policy from resources.",
        policy_arn=policy_arn,
        **log_params,
    )
    await asyncio.gather(*tasks)
    await boto_crud_call(iam_client.delete_policy, PolicyArn=policy_arn)


async def apply_update_managed_policy(
    iam_client,
    policy_arn: str,
    template_policy_document: dict,
    existing_policy_document: dict,
    log_params: dict,
) -> list[ProposedChange]:
    response = []
    if isinstance(existing_policy_document, str):
        existing_policy_document = json.loads(existing_policy_document)
    policy_drift = await aio_wrapper(
        DeepDiff,
        existing_policy_document,
        template_policy_document,
        ignore_order=True,
        report_repetition=True,
    )
    # DeepDiff will return type changes as actual type functions and not strings,
    # and this will cause json serialization to fail later on when we process
    # the proposed changes. We force type changes to strings here.

    if policy_drift:
        policy_drift = json.loads(policy_drift.to_json())
        log_str = "Changes to the PolicyDocument discovered."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.UPDATE,
                attribute="policy_document",
                change_summary=policy_drift,
                resource_type="aws:policy_document",
                resource_id=policy_arn,
                current_value=existing_policy_document,
                new_value=template_policy_document,
            )
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            apply_awaitable = new_policy_version(
                iam_client,
                policy_arn,
                template_policy_document,
                policy_drift,
                log_str,
                log_params,
            )
            return await plugin_apply_wrapper(apply_awaitable, proposed_changes)

        log.debug(log_str, **log_params)
    return response


async def new_policy_version(
    iam_client, policy_arn, template_policy_document, policy_drift, log_str, log_params
):
    if policy_drift:
        policy_versions = await list_managed_policy_versions(iam_client, policy_arn)
        if len(policy_versions) == 5:
            await boto_crud_call(
                iam_client.delete_policy_version,
                PolicyArn=policy_arn,
                VersionId=get_oldest_policy_version_id(policy_versions),
            )

    log_str = f"{log_str} Updating PolicyDocument..."
    await boto_crud_call(
        iam_client.create_policy_version,
        PolicyArn=policy_arn,
        PolicyDocument=json.dumps(template_policy_document),
    )
    log.debug(log_str, **log_params)


async def apply_managed_policy_tags(
    iam_client,
    policy_arn: str,
    template_tags: list[dict],
    existing_tags: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    existing_tag_map = {tag["Key"]: tag.get("Value") for tag in existing_tags}
    template_tag_map = {tag["Key"]: tag.get("Value") for tag in template_tags}
    tags_to_apply = [
        tag
        for tag in template_tags
        if tag.get("Value") != existing_tag_map.get(tag["Key"])
    ]
    tasks = []
    response = []

    if tags_to_remove := [
        tag["Key"] for tag in existing_tags if not template_tag_map.get(tag["Key"])
    ]:
        log_str = "Stale tags discovered."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                attribute="tags",
                resource_type="aws:policy_document",
                resource_id=policy_arn,
                change_summary={"TagKeys": tags_to_remove},
            )
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Removing tags..."
            apply_awaitable = boto_crud_call(
                iam_client.untag_policy,
                PolicyArn=policy_arn,
                TagKeys=tags_to_remove,
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, tags=tags_to_remove, **log_params)

    if tags_to_apply:
        log_str = "New tags discovered in AWS."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_type="aws:policy_document",
                resource_id=policy_arn,
                attribute="tags",
                new_value=tag,
            )
            for tag in tags_to_apply
        ]
        response.extend(proposed_changes)
        if ctx.execute:
            log_str = f"{log_str} Adding tags..."
            apply_awaitable = boto_crud_call(
                iam_client.tag_policy, PolicyArn=policy_arn, Tags=tags_to_apply
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, tags=tags_to_apply, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def get_managed_policy_across_accounts(
    aws_accounts: list[AWSAccount], managed_policy_path: str, managed_policy_name: str
) -> dict:
    async def get_managed_policy_for_account(aws_account: AWSAccount):
        iam_client = await aws_account.get_boto3_client("iam")
        arn = f"arn:aws:iam::{aws_account.account_id}:policy{managed_policy_path}{managed_policy_name}"
        return {aws_account.account_id: await get_managed_policy(iam_client, arn)}

    account_on_managed_policies = await asyncio.gather(
        *[get_managed_policy_for_account(aws_account) for aws_account in aws_accounts]
    )
    return {
        account_id: mp
        for resp in account_on_managed_policies
        for account_id, mp in resp.items()
    }



from __future__ import annotations

from typing import Optional

from pydantic import Field, constr

from iambic.core.models import ExpiryModel
from iambic.plugins.v0_1_0.aws.models import ARN_RE, AccessModel


class Path(AccessModel):
    file_path: str = Field(..., hidden_from_schema=True)

    @property
    def resource_type(self) -> str:
        return "iam:path"

    @property
    def resource_id(self) -> str:
        return self.file_path


class MaxSessionDuration(AccessModel):
    max_session_duration: int

    @property
    def resource_type(self) -> str:
        return "iam:max_session_duration"

    @property
    def resource_id(self) -> str:
        return str(self.max_session_duration)


class PermissionBoundary(ExpiryModel, AccessModel):
    # alias allows easily deserialize from boto3 response
    policy_arn: constr(regex=ARN_RE) = Field(alias="permissions_boundary_arn")

    # permissions_boundary_type to make it easy to deserialize from boto3 response
    # we won't actually serialize it back
    permissions_boundary_type: Optional[str] = Field(exclude=True)

    @property
    def resource_type(self):
        return "aws:iam:permission_boundary"

    @property
    def resource_id(self):
        return self.policy_arn
